from __future__ import annotations

import os
from typing import TYPE_CHECKING
from unittest.mock import ANY, MagicMock, patch

import pytest

from dissect.target.helpers.nfs.client.mount import Client as MountClient
from dissect.target.helpers.nfs.client.nfs import Client as NfsClient
from dissect.target.helpers.nfs.client.nfs import ReadDirResult
from dissect.target.helpers.nfs.nfs3 import (
    EntryPlus,
    FileAttributes,
    FileHandle,
    FileType,
    LookupResult,
    MountOK,
    NfsTime,
    ReadFileProc,
    ReadParams,
    SpecData,
)
from dissect.target.helpers.sunrpc.client import Client, auth_unix

if TYPE_CHECKING:
    from collections.abc import Iterator


@pytest.fixture
def mock_socket() -> Iterator[MagicMock]:
    with patch("socket.socket") as mock_socket:
        yield mock_socket


@pytest.fixture
def rpc_client() -> MagicMock:
    return MagicMock(spec=Client)


def test_portmap_call(mock_socket: MagicMock) -> None:
    portmap_request = b"\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x02\x00\x01\x86\xa0\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x86\xa3\x00\x00\x00\x03\x00\x00\x00\x06\x00\x00\x00\x00"  # noqa: E501

    portmap_response = b"\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x08\x01"  # noqa: E501

    # Mock the socket instance
    mock_sock_instance = MagicMock()
    mock_socket.return_value = mock_sock_instance

    client = Client.connect_port_mapper("localhost")

    # Set up the mock to return the response payload
    response_fragment_header = (len(portmap_response) | 0x80000000).to_bytes(4, "big")
    mock_sock_instance.recv.side_effect = [response_fragment_header, portmap_response]

    result = client.query_port_mapping(program=100003, version=3)

    # Verify that the request payload was sent
    portmap_request_fragment_header = (len(portmap_request) | 0x80000000).to_bytes(4, "big")
    mock_sock_instance.sendall.assert_called_with(portmap_request_fragment_header + portmap_request)

    # Verify that the result of the call equals the portmap_result variable
    assert result == 2049


def test_mount_call(mock_socket: MagicMock) -> None:
    mount_request = b"\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x02\x00\x01\x86\xa5\x00\x00\x00\x03\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\n/home/roel\x00\x00"  # noqa: E501

    mount_response = b"\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x01\x00\x07\x00\x02\x00\xec\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb4\x00\x00\x00\x01\x00\x00\x00\x01"  # noqa: E501

    # Mock the socket instance
    mock_sock_instance = MagicMock()
    mock_socket.return_value = mock_sock_instance

    # Set up the mock to return the response payload
    mount_response_fragment_header = (len(mount_response) | 0x80000000).to_bytes(4, "big")
    mock_sock_instance.recv.side_effect = [mount_response_fragment_header, mount_response]

    mount_client = MountClient.connect("localhost", 2049)
    result = mount_client.mount("/home/roel")

    # Verify that the request payload was sent
    portmap_request_fragment_header = (len(mount_request) | 0x80000000).to_bytes(4, "big")
    mock_sock_instance.sendall.assert_called_with(portmap_request_fragment_header + mount_request)

    # Verify that the result of the call equals the mount_result variable
    assert result == MountOK(
        filehandle=FileHandle(
            opaque=b"\x01\x00\x07\x00\x02\x00\xec\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb4"
        ),
        auth_flavors=[1],
    )


def test_readdir(mock_socket: MagicMock) -> None:
    readdir_request = b"\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x02\x00\x01\x86\xa3\x00\x00\x00\x03\x00\x00\x00\x11\x00\x00\x00\x01\x00\x00\x00\x1cq\xd5\x93D\x00\x00\x00\x07twigtop\x00\x00\x00\x03\xe8\x00\x00\x03\xe8\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x01\x00\x07\x00\xda&\xee\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb4\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x80\x00"  # noqa: E501
    readdir_response = b"\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x02\x00\x00\x01\xfd\x00\x00\x00\x02\x00\x00\x03\xe8\x00\x00\x03\xe8\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00Yq\x99zI\x1e5\r\x00\x00\x00\x00\x02\xee&\xdag\x8ar\xba)\xd7\xba.g\x8ar\x96\x18\xd7\x91<g\x8ar\x96\x18\xd7\x91<\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x02\xee\x84\x19\x00\x00\x00\x08test.txt>z;\x99\x07@\x9c_\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x01\xb4\x00\x00\x00\x01\x00\x00\x03\xe8\x00\x00\x03\xe8\x00\x00\x00\x00\x00\x00\x00\x05\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00Yq\x99zI\x1e5\r\x00\x00\x00\x00\x02\xee\x84\x19g\x8aqk\r\xa4\xb7\x8eg\x8aqg\x11\x93h\xf9g\x8aqg\x11\x93h\xf9\x00\x00\x00\x01\x00\x00\x00$\x01\x00\x07\x01\xda&\xee\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb4\x19\x84\xee\x02\xc1\x8a\x8c\\\x00\x00\x00\x01\x00\x00\x00\x00\x02\xee&\xda\x00\x00\x00\x01.\x00\x00\x00CjR\xafoN\x82\xf0\x00\x00\x00\x01\x00\x00\x00\x02\x00\x00\x01\xfd\x00\x00\x00\x02\x00\x00\x03\xe8\x00\x00\x03\xe8\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00Yq\x99zI\x1e5\r\x00\x00\x00\x00\x02\xee&\xdag\x8ar\xba)\xd7\xba.g\x8ar\x96\x18\xd7\x91<g\x8ar\x96\x18\xd7\x91<\x00\x00\x00\x01\x00\x00\x00\x1c\x01\x00\x07\x00\xda&\xee\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb4\x00\x00\x00\x01\x00\x00\x00\x00\x02\xee\x84\x07\x00\x00\x00\ttest2.txt\x00\x00\x00hF\x10\xd4\xd7u\xe2>\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x01\xb4\x00\x00\x00\x01\x00\x00\x03\xe8\x00\x00\x03\xe8\x00\x00\x00\x00\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00Yq\x99zI\x1e5\r\x00\x00\x00\x00\x02\xee\x84\x07g\x8ar\x96\x18\xd7\x91<g\x8ar\x9e+\x9e\xde\ng\x8ar\x9e+\x9e\xde\n\x00\x00\x00\x01\x00\x00\x00$\x01\x00\x07\x01\xda&\xee\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb4\x07\x84\xee\x02\x9524*\x00\x00\x00\x01\x00\x00\x00\x00\x02\xec\x00\x02\x00\x00\x00\x02..\x00\x00\x7f\xff\xff\xff\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01"  # noqa: E501

    # Mock the socket instance
    mock_sock_instance = MagicMock()
    mock_socket.return_value = mock_sock_instance

    # Set up the mock to return the response payload
    readdir_response_fragment_header = (len(readdir_response) | 0x80000000).to_bytes(4, "big")
    mock_sock_instance.recv.side_effect = [readdir_response_fragment_header, readdir_response]

    auth = auth_unix("twigtop", 1000, 1000, [])
    auth.credentials.stamp = 1909822276
    nfs_client = NfsClient.connect("localhost", 2049, auth, 666)
    result = nfs_client.readdir(
        FileHandle(opaque=b"\x01\x00\x07\x00\xda&\xee\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb4")
    )

    # Verify that the request payload was sent
    readdir_request_header = (len(readdir_request) | 0x80000000).to_bytes(4, "big")
    mock_sock_instance.sendall.assert_called_with(readdir_request_header + readdir_request)

    # Verify that the result of the call equals the readdir_result variable
    assert result == ReadDirResult(
        dir_attributes=FileAttributes(
            type=FileType.DIR,
            mode=509,
            nlink=2,
            uid=1000,
            gid=1000,
            size=4096,
            used=4096,
            rdev=SpecData(specdata1=0, specdata2=0),
            fsid=6445101292235666701,
            fileid=49161946,
            atime=NfsTime(seconds=1737126586, nseconds=702003758),
            mtime=NfsTime(seconds=1737126550, nseconds=416780604),
            ctime=NfsTime(seconds=1737126550, nseconds=416780604),
        ),
        entries=[
            EntryPlus(
                fileid=49185817,
                name="test.txt",
                cookie=4501976305947941983,
                attributes=FileAttributes(
                    type=FileType.REG,
                    mode=436,
                    nlink=1,
                    uid=1000,
                    gid=1000,
                    size=5,
                    used=4096,
                    rdev=SpecData(specdata1=0, specdata2=0),
                    fsid=6445101292235666701,
                    fileid=49185817,
                    atime=NfsTime(seconds=1737126251, nseconds=228898702),
                    mtime=NfsTime(seconds=1737126247, nseconds=294873337),
                    ctime=NfsTime(seconds=1737126247, nseconds=294873337),
                ),
                handle=FileHandle(
                    opaque=b"\x01\x00\x07\x01\xda&\xee\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb4\x19\x84\xee\x02\xc1\x8a\x8c\\"
                ),
            ),
            EntryPlus(
                fileid=49161946,
                name=".",
                cookie=4857786061512671984,
                attributes=FileAttributes(
                    type=FileType.DIR,
                    mode=509,
                    nlink=2,
                    uid=1000,
                    gid=1000,
                    size=4096,
                    used=4096,
                    rdev=SpecData(specdata1=0, specdata2=0),
                    fsid=6445101292235666701,
                    fileid=49161946,
                    atime=NfsTime(seconds=1737126586, nseconds=702003758),
                    mtime=NfsTime(seconds=1737126550, nseconds=416780604),
                    ctime=NfsTime(seconds=1737126550, nseconds=416780604),
                ),
                handle=FileHandle(
                    opaque=b"\x01\x00\x07\x00\xda&\xee\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb4"
                ),
            ),
            EntryPlus(
                fileid=49185799,
                name="test2.txt",
                cookie=7513711534648189502,
                attributes=FileAttributes(
                    type=FileType.REG,
                    mode=436,
                    nlink=1,
                    uid=1000,
                    gid=1000,
                    size=6,
                    used=4096,
                    rdev=SpecData(specdata1=0, specdata2=0),
                    fsid=6445101292235666701,
                    fileid=49185799,
                    atime=NfsTime(seconds=1737126550, nseconds=416780604),
                    mtime=NfsTime(seconds=1737126558, nseconds=731831818),
                    ctime=NfsTime(seconds=1737126558, nseconds=731831818),
                ),
                handle=FileHandle(
                    opaque=b"\x01\x00\x07\x01\xda&\xee\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb4\x07\x84\xee\x02\x9524*"
                ),
            ),
            EntryPlus(fileid=49020930, name="..", cookie=9223372036854775807, attributes=None, handle=None),
        ],
    )


def test_lookup(mock_socket: MagicMock) -> None:
    lookup_request = b"\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x02\x00\x01\x86\xa3\x00\x00\x00\x03\x00\x00\x00\x03\x00\x00\x00\x01\x00\x00\x00\x1ch \x9b-\x00\x00\x00\x07machine\x00\x00\x00\x03\xe8\x00\x00\x03\xe8\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x01\x00\x07\x00\xda&\xee\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb4\x00\x00\x00\x04dir1"  # noqa: E501
    lookup_response = b"\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00$\x01\x00\x07\x01\xda&\xee\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb41\x12\xf2\x02\x8f\x9958\x00\x00\x00\x01\x00\x00\x00\x02\x00\x00\x01\xfd\x00\x00\x00\x03\x00\x00\x03\xe8\x00\x00\x03\xe8\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00Yq\x99zI\x1e5\r\x00\x00\x00\x00\x02\xf2\x121g\xabzW9\x02\xc4\xf1g\xa9\xbc\xa3.\x95\xbf\x03g\xa9\xbc\xa3.\x95\xbf\x03\x00\x00\x00\x01\x00\x00\x00\x02\x00\x00\x01\xf8\x00\x00\x00\x03\x00\x00\x03\xe8\x00\x00\x03\xe8\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00Yq\x99zI\x1e5\r\x00\x00\x00\x00\x02\xee&\xdag\xab}\xdb#R\x98\xdeg\xa9\xe05\x142Q+g\xab{\xdc)8\xfa\xf2"  # noqa: E501

    # Mock the socket instance
    mock_sock_instance = MagicMock()
    mock_socket.return_value = mock_sock_instance

    # Set up the mock to return the response payload
    readdir_response_fragment_header = (len(lookup_response) | 0x80000000).to_bytes(4, "big")
    mock_sock_instance.recv.side_effect = [readdir_response_fragment_header, lookup_response]

    auth = auth_unix("machine", 1000, 1000, [])
    auth.credentials.stamp = 1746967341
    nfs_client = NfsClient.connect("localhost", 2049, auth, 666)
    parent_handle = FileHandle(
        opaque=b"\x01\x00\x07\x00\xda&\xee\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb4"
    )
    result = nfs_client.lookup("dir1", parent_handle)

    # Verify that the request payload was sent
    readdir_request_header = (len(lookup_request) | 0x80000000).to_bytes(4, "big")
    mock_sock_instance.sendall.assert_called_with(readdir_request_header + lookup_request)

    # Verify that the result of the call equals the result
    assert result == LookupResult(
        object=FileHandle(
            opaque=b"\x01\x00\x07\x01\xda&\xee\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb41\x12\xf2\x02\x8f\x9958"
        ),
        obj_attributes=FileAttributes(
            type=FileType.DIR,
            mode=509,
            nlink=3,
            uid=1000,
            gid=1000,
            size=4096,
            used=4096,
            rdev=SpecData(specdata1=0, specdata2=0),
            fsid=6445101292235666701,
            fileid=49418801,
            atime=NfsTime(seconds=1739291223, nseconds=956482801),
            mtime=NfsTime(seconds=1739177123, nseconds=781565699),
            ctime=NfsTime(seconds=1739177123, nseconds=781565699),
        ),
        dir_attributes=FileAttributes(
            type=FileType.DIR,
            mode=504,
            nlink=3,
            uid=1000,
            gid=1000,
            size=4096,
            used=4096,
            rdev=SpecData(specdata1=0, specdata2=0),
            fsid=6445101292235666701,
            fileid=49161946,
            atime=NfsTime(seconds=1739292123, nseconds=592615646),
            mtime=NfsTime(seconds=1739186229, nseconds=338841899),
            ctime=NfsTime(seconds=1739291612, nseconds=691600114),
        ),
    )


def test_getattr(mock_socket: MagicMock) -> None:
    getattr_request = b"\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x02\x00\x01\x86\xa3\x00\x00\x00\x03\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x1c\xe3\xee-G\x00\x00\x00\x07machine\x00\x00\x00\x03\xe8\x00\x00\x03\xe8\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x01\x00\x07\x00\xda&\xee\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb4"  # noqa: E501
    getattr_response = b"\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x01\xf8\x00\x00\x00\x03\x00\x00\x03\xe8\x00\x00\x03\xe8\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00Yq\x99zI\x1e5\r\x00\x00\x00\x00\x02\xee&\xdag\xad\xae\xf5\x1f\x12\xc8\xecg\xa9\xe05\x142Q+g\xab{\xdc)8\xfa\xf2"  # noqa: E501

    # Mock the socket instance
    mock_sock_instance = MagicMock()
    mock_socket.return_value = mock_sock_instance

    # Set up the mock to return the response payload
    readdir_response_fragment_header = (len(getattr_response) | 0x80000000).to_bytes(4, "big")
    mock_sock_instance.recv.side_effect = [readdir_response_fragment_header, getattr_response]

    auth = auth_unix("machine", 1000, 1000, [])
    auth.credentials.stamp = 3824037191
    nfs_client = NfsClient.connect("localhost", 2049, auth, 666)
    file_handle = FileHandle(
        opaque=b"\x01\x00\x07\x00\xda&\xee\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb4"
    )
    result = nfs_client.getattr(file_handle)

    # Verify that the request payload was sent
    readdir_request_header = (len(getattr_request) | 0x80000000).to_bytes(4, "big")
    mock_sock_instance.sendall.assert_called_with(readdir_request_header + getattr_request)

    # Verify that the result of the call equals the result
    assert result == FileAttributes(
        type=FileType.DIR,
        mode=504,
        nlink=3,
        uid=1000,
        gid=1000,
        size=4096,
        used=4096,
        rdev=SpecData(specdata1=0, specdata2=0),
        fsid=6445101292235666701,
        fileid=49161946,
        atime=NfsTime(seconds=1739435765, nseconds=521324780),
        mtime=NfsTime(seconds=1739186229, nseconds=338841899),
        ctime=NfsTime(seconds=1739291612, nseconds=691600114),
    )


def test_readlink(mock_socket: MagicMock) -> None:
    readlink_request = b"\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x02\x00\x01\x86\xa3\x00\x00\x00\x03\x00\x00\x00\x05\x00\x00\x00\x01\x00\x00\x00\x1ccy7\xba\x00\x00\x00\x07machine\x00\x00\x00\x03\xe8\x00\x00\x03\xe8\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00,\x01\x00\x07\x02\xda&\xee\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb4\x0f{\xee\x02\xefoz\xef\xda&\xee\x02'/\x00\x91"  # noqa: E501
    readlink_response = b"\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x05\x00\x00\x01\xff\x00\x00\x00\x01\x00\x00\x03\xe8\x00\x00\x03\xe8\x00\x00\x00\x00\x00\x00\x00\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00Yq\x99zI\x1e5\r\x00\x00\x00\x00\x02\xee{\x0fg\xac\xba\xc3\r5\xab\xa0g\xa9\xbe\x0c:H\x15\xc0g\xa9\xbe\x0c:H\x15\xc0\x00\x00\x00\x0edir1/dir2/dir3\x00\x00"  # noqa: E501

    # Mock the socket instance
    mock_sock_instance = MagicMock()
    mock_socket.return_value = mock_sock_instance

    # Set up the mock to return the response payload
    readdir_response_fragment_header = (len(readlink_response) | 0x80000000).to_bytes(4, "big")
    mock_sock_instance.recv.side_effect = [readdir_response_fragment_header, readlink_response]

    auth = auth_unix("machine", 1000, 1000, [])
    auth.credentials.stamp = 1668888506
    nfs_client = NfsClient.connect("localhost", 2049, auth, 666)
    file_handle = FileHandle(
        opaque=b"\x01\x00\x07\x02\xda&\xee\x02\x00\x00\x00\x00\xb5g\x131&\xf1I\xed\xb8R\rx\\h8\xb4\x0f{\xee\x02\xefoz\xef\xda&\xee\x02'/\x00\x91"
    )
    result = nfs_client.readlink(file_handle)

    # Verify that the request payload was sent
    readdir_request_header = (len(readlink_request) | 0x80000000).to_bytes(4, "big")
    mock_sock_instance.sendall.assert_called_with(readdir_request_header + readlink_request)

    # Verify that the result of the call equals the result
    assert result == "dir1/dir2/dir3"


def test_readfile(rpc_client: MagicMock) -> None:
    nfs_client = NfsClient(rpc_client)
    file_handle = FileHandle(opaque=b"file_handle")

    # Generate random binary data of 2.5 times the READ_CHUNK_SIZE
    data_size = int(2.5 * NfsClient.READ_CHUNK_SIZE)
    random_data = os.urandom(data_size)

    # Mock the responses to return the relevant chunks
    chunks = [random_data[i : i + NfsClient.READ_CHUNK_SIZE] for i in range(0, data_size, NfsClient.READ_CHUNK_SIZE)]
    responses = []
    for i, chunk in enumerate(chunks):
        eof = i == len(chunks) - 1
        response = MagicMock()
        response.data = chunk
        response.count = len(chunk)
        response.eof = eof
        responses.append(response)

    rpc_client.call.side_effect = responses

    # Read the file using the readfile method
    received_data = b"".join(nfs_client.readfile(file_handle))

    # Compare the received file with the generated one
    assert received_data == random_data
    assert rpc_client.call.call_count == len(chunks)
    for i, _ in enumerate(chunks):
        offset = i * NfsClient.READ_CHUNK_SIZE
        rpc_client.call.assert_any_call(
            ReadFileProc, ReadParams(file_handle, offset, NfsClient.READ_CHUNK_SIZE), ANY, ANY
        )



from __future__ import annotations

import json
import pathlib
from typing import TYPE_CHECKING, Any
from unittest.mock import patch

import pytest

from dissect.target.plugins.apps.webserver import iis
from dissect.target.plugins.os.windows import amcache
from dissect.target.tools.dump import (
    COMPRESSION_TO_EXT,
    SERIALIZERS,
    STATE_FILE_NAME,
    Compression,
    Serialization,
    create_state,
    execute_pipeline,
)
from dissect.target.tools.dump import main as target_dump
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.fixture
def target_win_iis_amcache(target_win_tzinfo: Target, fs_win: VirtualFilesystem) -> iis.Target:
    config_path = absolute_path("_data/plugins/apps/webserver/iis/iis-applicationHost-iis.config")
    data_dir = absolute_path("_data/plugins/apps/webserver/iis/iis-logs-iis")

    fs_win.map_file("windows/system32/inetsrv/config/applicationHost.config", config_path)
    fs_win.map_dir("Users/John/iis-logs", data_dir)

    target_win_tzinfo.add_plugin(iis.IISLogsPlugin)

    amcache_file = absolute_path("_data/plugins/os/windows/amcache/amcache-new.hve")
    fs_win.map_file("windows/appcompat/programs/amcache.hve", amcache_file)

    target_win_tzinfo.add_plugin(amcache.AmcachePlugin)
    return target_win_tzinfo


@pytest.mark.parametrize(
    ("serialization_name", "compression_name"),
    [
        ("jsonlines", "gzip"),
        ("jsonlines", "bzip2"),
        ("jsonlines", None),
        ("msgpack", "gzip"),
        ("msgpack", "bzip2"),
        ("msgpack", None),
    ],
)
def test_execute_pipeline(
    serialization_name: str,
    compression_name: str | None,
    target_win_iis_amcache: Target,
    tmp_path: pathlib.Path,
) -> None:
    def mock_get_targets(targets: list[str]) -> Iterator[Target]:
        yield target_win_iis_amcache

    def mock_log_progress(stream: Iterable[Any]) -> Iterator[Any]:
        count = 0
        for element in stream:
            yield element
            count += 1
        assert count == 128

    with (
        patch("dissect.target.tools.dump.log_progress", new=mock_log_progress),
    ):
        output_dir = tmp_path / "output"

        serialization = Serialization(serialization_name)
        compression = Compression(compression_name) if compression_name else Compression.NONE

        functions = "iis.logs,amcache.applications"
        _state = create_state(
            output_dir=output_dir,
            target_paths=["dummy"],
            functions=functions,
            excluded_functions=[],
            serialization=serialization,
            compression=compression,
        )

        execute_pipeline(
            state=_state,
            targets=mock_get_targets(["dummy"]),
            arguments=[],
            dry_run=False,
            limit=None,
        )

        target_name = target_win_iis_amcache.name

        # verify that iis records are in place

        assert (output_dir / target_name / "iis").exists()

        serialization_ext = SERIALIZERS[serialization]["ext"]
        compression_ext = COMPRESSION_TO_EXT[compression]

        iis_sink_filename = f"filesystem_windows_iis_logs.{serialization_ext}"
        if compression_ext:
            iis_sink_filename += f".{compression_ext}"

        assert (output_dir / target_name / "iis" / iis_sink_filename).exists()

        # verify that amcache.applications records are in place

        assert (output_dir / target_name / "amcache").exists()

        amcache_sink_filename = f"windows_appcompat_InventoryApplication.{serialization_ext}"
        if compression_ext:
            amcache_sink_filename += f".{compression_ext}"

        assert (output_dir / target_name / "amcache" / amcache_sink_filename).exists()

        # verify that serialized state is in place
        state_path = output_dir / STATE_FILE_NAME
        assert state_path.exists()

        state_blob = json.loads(state_path.read_text())

        assert state_blob["compression"] == str(compression_name)
        assert state_blob["serialization"] == serialization_name
        assert state_blob["target_paths"] == ["dummy"]
        assert state_blob["functions"] == functions

        sink_blobs = state_blob["sinks"]

        assert len(sink_blobs) == 2
        assert {s["func"] for s in sink_blobs} == set(functions.split(","))

        assert all(s["is_dirty"] is False for s in sink_blobs)

        # validate iis sink blob
        iis_sink_blob = next(s for s in state_blob["sinks"] if s["func"] == "iis.logs")
        assert iis_sink_blob["record_count"] == 10
        assert iis_sink_blob["target_path"] == str(target_win_iis_amcache.path)
        assert iis_sink_blob["path"] == str(pathlib.Path(target_name) / "iis" / iis_sink_filename)

        # validate amcache sink blob
        amcache_sink_blob = next(s for s in state_blob["sinks"] if s["func"] == "amcache.applications")
        assert amcache_sink_blob["record_count"] == 118
        assert amcache_sink_blob["target_path"] == str(target_win_iis_amcache.path)
        assert amcache_sink_blob["path"] == str(pathlib.Path(target_name) / "amcache" / amcache_sink_filename)


@pytest.mark.parametrize("limit", [5, 15, None])
def test_execute_pipeline_limited(limit: int | None, target_win_iis_amcache: Target, tmp_path: pathlib.Path) -> None:
    def mock_get_targets(targets: list[str]) -> Iterator[Target]:
        yield target_win_iis_amcache

    def mock_log_progress(stream: Iterable[Any]) -> Iterator[Any]:
        count = 0
        for element in stream:
            yield element
            count += 1

        if limit:
            assert count == limit
        else:
            assert count == 128

    with (
        patch("dissect.target.tools.dump.get_targets", new=mock_get_targets),
        patch("dissect.target.tools.dump.log_progress", new=mock_log_progress),
    ):
        output_dir = tmp_path / "output"

        functions = "iis.logs,amcache.applications"
        _state = create_state(
            output_dir=output_dir,
            target_paths=["dummy"],
            functions=functions,
            excluded_functions=[],
            serialization=Serialization.JSONLINES,
            compression=Compression.NONE,
        )

        execute_pipeline(
            state=_state,
            targets=mock_get_targets(["dummy"]),
            arguments=[],
            dry_run=False,
            limit=limit,
        )

        target_name = target_win_iis_amcache.name

        # verify that iis records are in place
        assert (output_dir / target_name / "iis").exists()

        serialization_ext = SERIALIZERS[Serialization.JSONLINES]["ext"]
        iis_sink_filename = f"filesystem_windows_iis_logs.{serialization_ext}"

        assert (output_dir / target_name / "iis" / iis_sink_filename).exists()

        # verify that serialized state is in place
        state_path = output_dir / STATE_FILE_NAME
        assert state_path.exists()

        state_blob = json.loads(state_path.read_text())

        assert state_blob["compression"] == "None"
        sink_blobs = state_blob["sinks"]

        iis_sink_blob = next(s for s in state_blob["sinks"] if s["func"] == "iis.logs")

        # verify that amcache.applications records are absent if `limit` is smaller/equal than 10 (the number
        # of iis records), and present if `limit` is `None` or larger than 10.
        if limit and limit <= 10:
            assert not (output_dir / target_name / "amcache").exists()

            assert len(sink_blobs) == 1
            assert {s["func"] for s in sink_blobs} == {"iis.logs"}

            assert iis_sink_blob["record_count"] == limit
            if limit < 10:
                # if limit cuts the record stream, the sink should be marked as dirty
                assert iis_sink_blob["is_dirty"]
            else:
                assert not iis_sink_blob["is_dirty"]

        else:
            assert (output_dir / target_name / "amcache").exists()
            amcache_sink_filename = f"windows_appcompat_InventoryApplication.{serialization_ext}"
            assert (output_dir / target_name / "amcache" / amcache_sink_filename).exists()

            assert len(sink_blobs) == 2
            assert {s["func"] for s in sink_blobs} == set(functions.split(","))

            amcache_sink_blob = next(s for s in state_blob["sinks"] if s["func"] == "amcache.applications")

            if limit:
                assert amcache_sink_blob["record_count"] == limit - 10
                assert amcache_sink_blob["is_dirty"]
            else:
                assert amcache_sink_blob["record_count"] == 118
                assert not amcache_sink_blob["is_dirty"]


def test_execute_excluded_plugins(target_win_iis_amcache: Target, tmp_path: pathlib.Path) -> None:
    def mock_get_targets(targets: list[str]) -> Iterator[Target]:
        yield target_win_iis_amcache

    with patch("dissect.target.tools.dump.get_targets", new=mock_get_targets):
        output_dir = tmp_path.joinpath("output")

        functions = "amcache.applications"
        _state = create_state(
            output_dir=output_dir,
            target_paths=["dummy"],
            functions=functions,
            excluded_functions=[functions],
            serialization=Serialization.JSONLINES,
            compression=Compression.NONE,
        )

        execute_pipeline(
            state=_state,
            targets=mock_get_targets(["dummy"]),
            arguments=[],
            dry_run=False,
            limit=None,
        )

        target_name = target_win_iis_amcache.name

        assert output_dir.joinpath(target_name).exists()
        assert not output_dir.joinpath(f"{target_name}/amcahce.applications").exists()


# tool tests
def test_dump(monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path) -> None:
    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            [
                "target-dump",
                "-f",
                "walkfs",
                str(absolute_path("_data/loaders/tar/test-archive.tar.gz")),
                "-o",
                str(tmp_path),
            ],
        )

        target_dump()

        assert tmp_path.joinpath("target-dump.state.json").exists()

        entry = tmp_path.joinpath("test-archive.tar.gz/walkfs/filesystem_entry.jsonl")
        assert entry.exists()
        assert "test-file.txt" in entry.read_text()


def test_dump_excluded_plugins(monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path) -> None:
    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            [
                "target-dump",
                "-f",
                "mft,walkfs",
                "-xf",
                "mft",
                str(absolute_path("_data/loaders/tar/test-archive.tar.gz")),
                "-o",
                str(tmp_path),
            ],
        )

        target_dump()

        assert tmp_path.joinpath("target-dump.state.json").exists()

        entry = tmp_path.joinpath("test-archive.tar.gz/walkfs/filesystem_entry.jsonl")
        assert entry.exists()
        assert "test-file.txt" in entry.read_text()

        assert not tmp_path.joinpath("test-archive.tar.gz/mft").exists()



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.filesystems.tar import TarFilesystem
from dissect.target.loaders.tar import TarSubLoader

if TYPE_CHECKING:
    import tarfile as tf
    from pathlib import Path

    from dissect.target.target import Target


class NsCollectorTarSubLoader(TarSubLoader):
    """Loader for tar-based Netscaler Techsupport Collector files.

    References:
        - https://developer-docs.netscaler.com/en-us/adc-command-reference-int/current-release/utility/techsupport.html
        - https://support.citrix.com/support-home/kbsearch/article?articleNumber=CTX222958
    """

    @staticmethod
    def detect(path: Path, tarfile: tf.TarFile) -> bool:
        if not (names := tarfile.getnames()):
            return False

        root = names[0]
        required_paths = {f"{root}/nsconfig/ns.conf", f"{root}/shell"}

        return required_paths.issubset(names)

    def map(self, target: Target) -> None:
        fs = TarFilesystem(self.tar.fileobj, base=self.tar.getnames()[0])
        target.filesystems.add(fs)

        # Symlink /nsconfig to /flash/nsconfig to make Citrix parsers compatible
        target.fs.symlink("/nsconfig", "/flash/nsconfig")



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.plugins.os.windows.wua_history import WuaHistoryPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_wua_history_plugin(target_win: Target, fs_win: VirtualFilesystem) -> None:
    wua_history_dir = absolute_path("_data/plugins/os/windows/wua_history/DataStore.edb.gz")
    fs_win.map_file("Windows/SoftwareDistribution/DataStore/DataStore.edb", wua_history_dir, compression="gzip")
    target_win.add_plugin(WuaHistoryPlugin)

    records = list(target_win.wua_history())
    record = records[0]

    assert len(records) == 80
    assert record.ts == datetime(2024, 7, 13, 16, 49, 9, 0, tzinfo=timezone.utc)
    assert record.id_event == 80
    assert record.status_id == 1
    assert record.status_mapped == "Success"
    assert record.server_selection == 2
    assert record.server_selection_mapped == "ssWindowsUpdate"
    assert record.mapped_result == "0x0"
    assert record.mapped_result_string == "Success"
    assert record.mapped_result_description == "Success"
    assert record.unmapped_result == "0x0"
    assert record.unmapped_result_string == "Success"
    assert record.unmapped_result_description == "Success"
    assert record.update_id == "71156214-2cb8-4c44-b11c-3ec1a0366f55"
    assert record.server_id == "9482f4b4-e343-43b6-b170-9a65bc822c77"
    assert record.server_id_mapped == "Windows Update"
    assert record.flags == 1
    assert record.client_id == "UpdateOrchestrator"
    assert (
        record.title
        == "Security Intelligence Update for Microsoft Defender Antivirus - KB2267602 (Version 1.415.74.0) - Current "
        "Channel (Broad)"
    )
    assert (
        record.description
        == "Install this update to revise the files that are used to detect viruses, spyware, and other potentially "
        "unwanted software. Once you have installed this item, it cannot be removed."
    )
    assert record.support_url == "https://go.microsoft.com/fwlink/?LinkId=52661"
    assert record.categories == ["Microsoft Defender Antivirus"]
    assert record.more_info_url == ["https://go.microsoft.com/fwlink/?linkid=2007160"]
    assert record.id_user == 0
    assert record.is_service_is_additional == "False"
    assert record.classification == "e0789628-ce08-4437-be74-2495b842f43b"
    assert record.classification_mapped == "DefinitionUpdates"
    assert record.kb == "KB2267602"



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.containerimage import ContainerImageTarSubLoader
from dissect.target.loaders.tar import TarLoader
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target]) -> None:
    """Test that we correctly use ``ContainerImageTarSubLoader`` when opening a ``Target``."""
    path = absolute_path("_data/loaders/containerimage/alpine-docker.tar")

    target = opener(path)
    assert isinstance(target._loader, TarLoader)
    assert isinstance(target._loader.subloader, ContainerImageTarSubLoader)
    assert target.path == path


def test_docker() -> None:
    """Test if we map a Docker image correctly."""
    path = absolute_path("_data/loaders/containerimage/alpine-docker.tar")

    loader = loader_open(path)
    assert isinstance(loader, TarLoader)

    t = Target()
    loader.map(t)
    assert isinstance(loader.subloader, ContainerImageTarSubLoader)

    assert loader.subloader.name == "alpine:latest"
    assert list(map(str, loader.subloader.layers)) == [
        "blobs/sha256/a0904247e36a7726c03c71ee48f3e64462021c88dafeb13f37fdaf613b27f11c"
    ]
    assert loader.subloader.config["created"] == "2025-01-08T12:07:30Z"

    assert len(t.filesystems) == 1
    assert len(t.filesystems[0].layers) == 3

    t.apply()
    assert sorted(map(str, t.fs.path("/").iterdir())) == [
        "/$fs$",
        "/bin",
        "/dev",
        "/etc",
        "/home",
        "/lib",
        "/media",
        "/mnt",
        "/opt",
        "/proc",
        "/root",
        "/run",
        "/sbin",
        "/srv",
        "/sys",
        "/tmp",
        "/usr",
        "/var",
    ]


def test_oci_podman() -> None:
    """Test if we map a Podman OCI image correctly."""
    path = absolute_path("_data/loaders/containerimage/alpine-oci.tar")

    loader = loader_open(path)
    assert isinstance(loader, TarLoader)

    t = Target()
    loader.map(t)
    assert isinstance(loader.subloader, ContainerImageTarSubLoader)

    assert loader.subloader.name == "docker.io/library/alpine:latest"
    assert list(map(str, loader.subloader.layers)) == [
        "08000c18d16dadf9553d747a58cf44023423a9ab010aab96cf263d2216b8b350.tar"
    ]
    assert loader.subloader.manifest == {
        "Config": "aded1e1a5b3705116fa0a92ba074a5e0b0031647d9c315983ccba2ee5428ec8b.json",
        "RepoTags": ["docker.io/library/alpine:latest"],
        "Layers": ["08000c18d16dadf9553d747a58cf44023423a9ab010aab96cf263d2216b8b350.tar"],
    }

    assert loader.subloader.config == {
        "architecture": "amd64",
        "config": {
            "Env": ["PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin"],
            "Cmd": ["/bin/sh"],
            "WorkingDir": "/",
        },
        "created": "2025-02-14T03:28:36Z",
        "history": [
            {
                "created": "2025-02-14T03:28:36Z",
                "created_by": "ADD alpine-minirootfs-3.21.3-x86_64.tar.gz / # buildkit",
                "comment": "buildkit.dockerfile.v0",
            },
            {
                "created": "2025-02-14T03:28:36Z",
                "created_by": 'CMD ["/bin/sh"]',
                "comment": "buildkit.dockerfile.v0",
                "empty_layer": True,
            },
        ],
        "os": "linux",
        "rootfs": {
            "type": "layers",
            "diff_ids": ["sha256:08000c18d16dadf9553d747a58cf44023423a9ab010aab96cf263d2216b8b350"],
        },
    }

    t.apply()

    assert sorted(map(str, t.fs.path("/").iterdir())) == [
        "/$fs$",
        "/bin",
        "/dev",
        "/etc",
        "/home",
        "/lib",
        "/media",
        "/mnt",
        "/opt",
        "/proc",
        "/root",
        "/run",
        "/sbin",
        "/srv",
        "/sys",
        "/tmp",
        "/usr",
        "/var",
    ]



from __future__ import annotations

import logging
from datetime import datetime, timezone
from typing import TYPE_CHECKING

import pytest
from flow.record.fieldtypes import datetime as dt

from dissect.target.plugins.os.unix.log.journal import JournalMessagePriority, JournalPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pytest_benchmark.fixture import BenchmarkFixture

    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_journal_plugin(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test linux systemd journal file parsing."""

    data_file = absolute_path("_data/plugins/os/unix/log/journal/journal")
    fs_unix.map_file("var/log/journal/1337/user-1000.journal", data_file)
    target_unix.add_plugin(JournalPlugin)

    results = list(target_unix.journal())
    assert len(results) == 2

    record = results[0]
    assert record.ts == dt("2023-05-19T16:22:38.841870+00:00")
    assert record.message == (
        "Window manager warning: last_user_time (928062) is greater than comparison timestamp (928031).  "
        "This most likely represents a buggy client sending inaccurate timestamps in messages such as "
        "_NET_ACTIVE_WINDOW.  Trying to work around..."
    )
    assert record.syslog_facility == "3"
    assert record.syslog_identifier == "gnome-shell"
    assert record.pid == 2096
    assert record.transport == "stdout"
    assert record.kernel_boot_id == "5ffb8dae595f4679988508fd45d9c8c4"
    assert record.priority == "info"
    assert record.source == "/var/log/journal/1337/user-1000.journal"


@pytest.mark.benchmark
def test_benchmark_journal(target_unix: Target, fs_unix: VirtualFilesystem, benchmark: BenchmarkFixture) -> None:
    """Test if we can parse some large journal files. this demonstrates how slow the journal plugin is."""

    system_journal = absolute_path("_data/plugins/os/unix/log/journal/system.journal")
    user_journal = absolute_path("_data/plugins/os/unix/log/journal/user-1000.journal")

    fs_unix.map_file("/var/log/journal/deadbeef/system.journal", system_journal)  # contains 252 entries
    fs_unix.map_file("/var/log/journal/deadbeef/user-1000.journal", user_journal)  # contains 17986 entries

    result = benchmark(lambda: next(JournalPlugin(target_unix).journal()))

    assert result.ts == datetime(2025, 1, 15, 14, 25, 1, 607354, tzinfo=timezone.utc)
    assert result.message == "pam_unix(cron:session): session opened for user root(uid=0) by root(uid=0)"


def test_journal_plugin_unused_object(
    caplog: pytest.LogCaptureFixture, target_unix: Target, fs_unix: VirtualFilesystem
) -> None:
    """Test if we can handle OBJECT_UNUSED in journal files correctly."""

    # unused.journal is a modified copy of system.journal at offset 0x393260.
    # the next_entry_array_offset was set from 0x00 to 0x3C1337.
    data_file = absolute_path("_data/plugins/os/unix/log/journal/unused.journal")
    fs_unix.map_file("/var/log/journal/deadbeef/system.journal", data_file)
    target_unix.add_plugin(JournalPlugin)

    with caplog.at_level(logging.WARNING, target_unix.log.name):
        results = list(target_unix.journal())

    assert "ObjectType OBJECT_UNUSED encountered for next OBJECT_ENTRY_ARRAY offset at 0x3C1337" in caplog.text
    assert len(results) == 252


@pytest.mark.parametrize(
    ("input", "expected_name", "expected_value"),
    [
        (0, "EMERG", 0),
        (1, "ALERT", 1),
        (2, "CRIT", 2),
        (3, "ERR", 3),
        (4, "WARNING", 4),
        (5, "NOTICE", 5),
        (6, "INFO", 6),
        (7, "DEBUG", 7),
        (1337, "1337", 1337),
    ],
)
def test_journal_plugin_priority_enum(input: int, expected_name: str, expected_value: int) -> None:
    """Test if we can handle different priority values."""
    priority = JournalMessagePriority(input)
    assert priority.name == expected_name
    assert priority.value == expected_value



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from flow.record.fieldtypes import datetime

from dissect.target.plugins.os.windows.lnk import LnkPlugin, LnkRecord
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("path", "error"),
    [
        pytest.param(None, False, id="No directory provided"),
        pytest.param("sysvol/users/", False, id="Directory provided"),
        pytest.param("sysvol/users/pestudio.lnk", False, id="File provided"),
        pytest.param("sysvol/users/non-existing.lnk", True, id="Non-existing file provided"),
    ],
)
def test_lnk(
    target_win: Target, fs_win: VirtualFilesystem, path: str | None, error: bool, caplog: pytest.LogCaptureFixture
) -> None:
    lnk_file = absolute_path("_data/plugins/os/windows/lnk/pestudio.lnk")
    fs_win.map_file("Users/pestudio.lnk", lnk_file)

    target_win.add_plugin(LnkPlugin)

    records = list(target_win.lnk(path))

    if error:
        assert caplog.records[-1].levelname == "ERROR"
        assert len(records) == 0
    else:
        assert len(records) == 1

        record = records[0]

        assert isinstance(record, type(LnkRecord()))
        assert str(record.lnk_path) == "sysvol\\users\\pestudio.lnk"
        assert record.lnk_name is None
        assert str(record.lnk_relativepath) == "pestudio.exe"
        assert str(record.lnk_workdir) == "C:\\Program Files\\pestudio"
        assert record.lnk_arguments is None
        assert record.lnk_iconlocation is None
        assert record.local_base_path == "C:\\Program Files\\pestudio\\pestudio.exe"
        assert record.common_path_suffix == ""
        assert record.lnk_net_name is None
        assert record.lnk_device_name is None
        assert str(record.lnk_full_path) == "C:\\Program Files\\pestudio\\pestudio.exe"
        assert record.machine_id == "desktop-i2purd1"
        assert record.target_mtime == datetime("2021-10-09T07:24:42+00:00")
        assert record.target_atime == datetime("2021-10-16T15:26:55.033125+00:00")
        assert record.target_ctime == datetime("2021-10-16T15:26:20.406921+00:00")



from __future__ import annotations

import stat
from typing import TYPE_CHECKING, BinaryIO

from dissect.util.stream import AlignedStream

from dissect.target.exceptions import FilesystemError, NotASymlinkError
from dissect.target.filesystem import (
    DirEntry,
    Filesystem,
    FilesystemEntry,
    VirtualDirectory,
    VirtualFile,
    VirtualFilesystem,
)
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.loaders.itunes import FileInfo, ITunesBackup


class ITunesFilesystem(Filesystem):
    """Filesystem implementation for iTunes backups."""

    __type__ = "itunes"

    def __init__(self, backup: ITunesBackup, *args, **kwargs):
        super().__init__(None, *args, **kwargs)
        self.backup = backup

        self._fs = VirtualFilesystem(alt_separator=self.alt_separator, case_sensitive=self.case_sensitive)

        for file in self.backup.files():
            entry_cls = (
                ITunesFilesystemDirectoryEntry if stat.S_IFMT(file.mode) == stat.S_IFDIR else ITunesFilesystemEntry
            )

            file_entry = entry_cls(self, file.translated_path, file)
            self._fs.map_file_entry(file.translated_path, file_entry)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        raise TypeError("Detect is not allowed on ITunesFilesystem class")

    def get(self, path: str, relentry: FilesystemEntry | None = None) -> FilesystemEntry:
        """Returns a ITunesFileEntry object corresponding to the given path."""
        return self._fs.get(path, relentry)


class ITunesFilesystemEntry(VirtualFile):
    fs: ITunesFilesystem
    entry: FileInfo

    def open(self) -> BinaryIO:
        """Returns file handle (file-like object)."""
        if self.is_dir():
            raise IsADirectoryError(self.path)

        if self.entry.backup.encrypted:
            return EncryptedFileStream(self.entry)
        return self.entry.get().open("rb")

    def scandir(self) -> Iterator[DirEntry]:
        if self.is_dir():
            return self._resolve().scandir()
        return super().scandir()

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a directory."""
        try:
            return stat.S_IFMT(self._resolve(follow_symlinks=follow_symlinks).entry.mode) == stat.S_IFDIR
        except FilesystemError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a file."""
        try:
            return stat.S_IFMT(self._resolve(follow_symlinks=follow_symlinks).entry.mode) == stat.S_IFREG
        except FilesystemError:
            return False

    def is_symlink(self) -> bool:
        """Return whether this entry is a link."""
        return stat.S_IFMT(self.entry.mode) == stat.S_IFLNK

    def readlink(self) -> str:
        """Read the link if this entry is a symlink. Returns a string."""
        if not self.is_symlink():
            raise NotASymlinkError
        return self.entry.metadata["Target"]

    def readlink_ext(self) -> FilesystemEntry:
        """Read the link if this entry is a symlink. Returns a filesystem entry."""
        # Can't use the one in VirtualFile as it overrides the FilesystemEntry
        return fsutil.resolve_link(self.fs, self.readlink(), self.path, alt_separator=self.fs.alt_separator)

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        """Return the stat information of this entry."""
        return self._resolve(follow_symlinks=follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        metadata = self.entry.metadata
        # ['mode', 'addr', 'dev', 'nlink', 'uid', 'gid', 'size', 'atime', 'mtime', 'ctime']
        return fsutil.stat_result(
            [
                metadata["Mode"],
                metadata["InodeNumber"],
                id(self.fs),
                1,
                metadata["UserID"],
                metadata["GroupID"],
                metadata["Size"],
                0,
                metadata["LastModified"],
                metadata["Birth"],
            ]
        )


class ITunesFilesystemDirectoryEntry(VirtualDirectory):
    def __init__(self, fs: Filesystem, path: str, entry: FileInfo):
        super().__init__(fs, path)
        self.entry = entry

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        """Return the stat information of this entry."""
        return self.lstat()

    def lstat(self) -> fsutil.stat_result:
        """Return the stat information of the given path, without resolving links."""
        metadata = self.entry.metadata
        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        return fsutil.stat_result(
            [
                metadata["Mode"],
                metadata["InodeNumber"],
                id(self.fs),
                0,
                metadata["UserID"],
                metadata["GroupID"],
                metadata["Size"],
                0,
                metadata["LastModified"],
                metadata["Birth"],
            ]
        )


class EncryptedFileStream(AlignedStream):
    """Transparently decrypted AES-CBC decrypted stream."""

    def __init__(self, file_info: FileInfo):
        super().__init__(file_info.size)
        self.file_info = file_info
        self.fh = file_info.get().open("rb")

        self.cipher = None
        self.cipher_offset = 0

        self._reset_cipher()

    def _reset_cipher(self) -> None:
        self.cipher = self.file_info.create_cipher()
        self.cipher_offset = 0

    def _seek_cipher(self, offset: int) -> None:
        """CBC is dependent on previous blocks so to seek the cipher, decrypt and discard to the wanted offset."""
        if offset < self.cipher_offset:
            self._reset_cipher()
            self.fh.seek(0)

        while self.cipher_offset < offset:
            read_size = min(offset - self.cipher_offset, self.align)
            self.cipher.decrypt(self.fh.read(read_size))
            self.cipher_offset += read_size

    def _read(self, offset: int, length: int) -> bytes:
        self._seek_cipher(offset)

        data = self.cipher.decrypt(self.fh.read(length))
        self.cipher_offset += length

        return data



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

import dissect.vmfs as vmfs
from dissect.vmfs.c_vmfs import c_vmfs

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import DirEntry, Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator


class VmfsFilesystem(Filesystem):
    __type__ = "vmfs"

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.vmfs = vmfs.VMFS(fh)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        """Detect a VMFS filesystem on a given file-like object."""
        fh.seek(c_vmfs.FS3_FS_HEADER_OFFSET)
        sector = fh.read(512)
        return int.from_bytes(sector[:4], "little") in (
            c_vmfs.VMFS_MAGIC_NUMBER,
            c_vmfs.VMFSL_MAGIC_NUMBER,
        )

    def get(self, path: str) -> FilesystemEntry:
        return VmfsFilesystemEntry(self, path, self._get_node(path))

    def _get_node(self, path: str, node: vmfs.FileDescriptor | None = None) -> vmfs.FileDescriptor:
        """Returns an internal VMFS entry for a given path and optional relative entry."""
        try:
            return self.vmfs.get(path, node)
        except vmfs.FileNotFoundError as e:
            raise FileNotFoundError(path) from e
        except vmfs.NotADirectoryError as e:
            raise NotADirectoryError(path) from e
        except vmfs.NotASymlinkError as e:
            raise NotASymlinkError(path) from e
        except vmfs.Error as e:
            raise FileNotFoundError(path) from e


class VmfsDirEntry(DirEntry):
    fs: VmfsFilesystem
    entry: vmfs.DirEntry

    def get(self) -> VmfsFilesystemEntry:
        return VmfsFilesystemEntry(self.fs, self.path, self.entry.file_descriptor)

    def is_dir(self, *, follow_symlinks: bool = True) -> bool:
        if follow_symlinks and self.is_symlink():
            return super().is_dir(follow_symlinks=follow_symlinks)

        return self.entry.type == c_vmfs.FS3_DescriptorType.DIRECTORY

    def is_file(self, *, follow_symlinks: bool = True) -> bool:
        if follow_symlinks and self.is_symlink():
            return super().is_file(follow_symlinks=follow_symlinks)

        return self.entry.type == c_vmfs.FS3_DescriptorType.REGFILE

    def is_symlink(self) -> bool:
        return self.entry.type == c_vmfs.FS3_DescriptorType.SYMLINK

    def stat(self, *, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.get().stat(follow_symlinks=follow_symlinks)


class VmfsFilesystemEntry(FilesystemEntry):
    fs: VmfsFilesystem
    entry: vmfs.FileDescriptor

    def get(self, path: str) -> FilesystemEntry:
        """Get a filesystem entry relative from the current one."""
        full_path = fsutil.join(self.path, path, alt_separator=self.fs.alt_separator)
        return VmfsFilesystemEntry(self.fs, full_path, self.fs._get_node(path, self.entry))

    def open(self) -> BinaryIO:
        """Returns file handle (file-like object)."""
        if self.is_dir():
            raise IsADirectoryError(self.path)
        return self._resolve().entry.open()

    def scandir(self) -> Iterator[VmfsDirEntry]:
        """List the directory contents of this directory. Returns a generator of filesystem entries."""
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        for entry in self._resolve().entry.iterdir():
            if entry.name in (".", ".."):
                continue

            yield VmfsDirEntry(self.fs, self.path, entry.name, entry)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a directory."""
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.is_dir()
        except FilesystemError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a file."""
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.is_file()
        except FilesystemError:
            return False

    def is_symlink(self) -> bool:
        """Return whether this entry is a link."""
        return self.entry.is_symlink()

    def readlink(self) -> str:
        """Read the link of the given path if it is a symlink. Returns a string."""
        if not self.is_symlink():
            raise NotASymlinkError

        return self.entry.link

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        """Return the stat information of this entry."""
        return self._resolve(follow_symlinks=follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        """Return the stat information of the given path, without resolving links."""
        meta = self.entry.metadata

        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        st_info = [
            self.entry.mode,
            self.entry.address,
            id(self.fs),
            meta.linkCount,
            meta.uid,
            meta.gid,
            self.entry.size,
            meta.atime,
            meta.mtime,
            meta.ctime,
        ]

        return fsutil.stat_result(st_info)



from __future__ import annotations

import math
from typing import TYPE_CHECKING, BinaryIO

import dissect.btrfs as btrfs
from dissect.btrfs.c_btrfs import c_btrfs

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import DirEntry, Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator


class BtrfsFilesystem(Filesystem):
    __type__ = "btrfs"
    __multi_volume__ = True

    def __init__(self, fh: BinaryIO | list[BinaryIO], *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.btrfs = btrfs.Btrfs(fh)
        self.subfs = self.open_subvolume()
        self.subvolume = self.subfs.subvolume

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        fh.seek(c_btrfs.BTRFS_SUPER_INFO_OFFSET)
        block = fh.read(4096)
        magic = int.from_bytes(block[64:72], "little")

        return magic == c_btrfs.BTRFS_MAGIC

    @staticmethod
    def _detect_id(fh: BinaryIO) -> bytes | None:
        # First field is csum, followed by fsid
        fh.seek(c_btrfs.BTRFS_SUPER_INFO_OFFSET + c_btrfs.BTRFS_CSUM_SIZE)
        return fh.read(c_btrfs.BTRFS_FSID_SIZE)

    def iter_subfs(self) -> Iterator[BtrfsSubvolumeFilesystem]:
        for subvol in self.btrfs.subvolumes():
            if subvol.objectid == self.subfs.subvolume.objectid:
                # Skip the default volume as it's already opened by the main filesystem
                continue
            yield self.open_subvolume(subvolid=subvol.objectid)

    def open_subvolume(self, subvol: str | None = None, subvolid: int | None = None) -> BtrfsSubvolumeFilesystem:
        return BtrfsSubvolumeFilesystem(self, subvol, subvolid)

    def get(self, path: str) -> FilesystemEntry:
        return self.subfs.get(path)


class BtrfsSubvolumeFilesystem(Filesystem):
    __type__ = "btrfs"

    def __init__(self, fs: BtrfsFilesystem, subvol: str | None = None, subvolid: int | None = None):
        super().__init__(fs.volume, alt_separator=fs.alt_separator, case_sensitive=fs.case_sensitive)
        if subvol is not None and subvolid is not None:
            raise ValueError("Only one of subvol or subvolid is allowed")

        self.fs = fs
        self.btrfs = fs.btrfs
        if subvol:
            self.subvolume = self.btrfs.find_subvolume(subvol)
        elif subvolid:
            self.subvolume = self.btrfs.open_subvolume(subvolid)
        else:
            self.subvolume = self.btrfs.default_subvolume

    def get(self, path: str) -> FilesystemEntry:
        return BtrfsFilesystemEntry(self, path, self._get_node(path))

    def _get_node(self, path: str, node: btrfs.INode | None = None) -> btrfs.INode:
        try:
            return self.subvolume.get(path, node)
        except btrfs.FileNotFoundError as e:
            raise FileNotFoundError(path) from e
        except btrfs.NotADirectoryError as e:
            raise NotADirectoryError(path) from e
        except btrfs.NotASymlinkError as e:
            raise NotASymlinkError(path) from e
        except btrfs.Error as e:
            raise FileNotFoundError(path) from e


class BtrfsDirEntry(DirEntry):
    fs: BtrfsSubvolumeFilesystem
    entry: btrfs.INode

    def get(self) -> BtrfsFilesystemEntry:
        return BtrfsFilesystemEntry(self.fs, self.path, self.entry)

    def stat(self, *, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.get().stat(follow_symlinks=follow_symlinks)


class BtrfsFilesystemEntry(FilesystemEntry):
    fs: BtrfsSubvolumeFilesystem
    entry: btrfs.INode

    def get(self, path: str) -> FilesystemEntry:
        path = fsutil.join(self.path, path, alt_separator=self.fs.alt_separator)
        return BtrfsFilesystemEntry(self.fs, path, self.fs._get_node(path, self.entry))

    def open(self) -> BinaryIO:
        if self.is_dir():
            raise IsADirectoryError(self.path)
        return self._resolve().entry.open()

    def scandir(self) -> Iterator[BtrfsDirEntry]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        for name, entry in self._resolve().entry.iterdir():
            if name in (".", ".."):
                continue

            # TODO: Separate INodes and DirEntry in dissect.btrfs
            yield BtrfsDirEntry(self.fs, self.path, name, entry)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.is_dir()
        except FilesystemError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.is_file()
        except FilesystemError:
            return False

    def is_symlink(self) -> bool:
        return self.entry.is_symlink()

    def readlink(self) -> str:
        if not self.is_symlink():
            raise NotASymlinkError

        return self.entry.link

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self._resolve(follow_symlinks=follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        entry = self.entry
        node = self.entry.inode

        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        st_info = fsutil.stat_result(
            [
                entry.mode,
                entry.inum,
                0,
                node.nlink,
                entry.uid,
                entry.gid,
                entry.size,
                # timestamp() returns a float which will fill both the integer and float fields
                entry.atime.timestamp(),
                entry.mtime.timestamp(),
                entry.ctime.timestamp(),
            ]
        )

        # Set the nanosecond resolution separately
        st_info.st_atime_ns = entry.atime_ns
        st_info.st_mtime_ns = entry.mtime_ns
        st_info.st_ctime_ns = entry.ctime_ns

        # Btrfs has a birth time, called otime
        st_info.st_birthtime = entry.otime.timestamp()
        st_info.st_birthtime_ns = entry.otime_ns

        # Add block information of the filesystem
        st_info.st_blksize = entry.btrfs.sector_size

        st_info.st_blocks = 0
        if not self.is_dir():
            st_info.st_blocks = (st_info.st_blksize // 512) * math.ceil(st_info.st_size / st_info.st_blksize)

        return st_info



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.locate.mlocate import MLocatePlugin, MLocateRecord
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_mlocate(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file("/var/lib/mlocate/mlocate.db", absolute_path("_data/plugins/os/unix/locate/mlocate.db"))
    target_unix.add_plugin(MLocatePlugin)

    records = list(target_unix.mlocate.locate())
    assert len(records) == 3317

    root_directory = records[0]
    assert isinstance(root_directory, type(MLocateRecord()))
    assert root_directory.parent.as_posix() == "/"
    assert root_directory.path.as_posix() == "/.dockerenv"
    assert root_directory.ts == datetime(1970, 1, 1, 0, 0, 0, tzinfo=timezone.utc)

    entry = records[1]
    assert isinstance(entry, type(MLocateRecord()))
    assert entry.parent.as_posix() == "/"
    assert entry.path.as_posix() == "/bin"
    assert entry.type == "file"  # symlink

    entry = records[2]
    assert isinstance(entry, type(MLocateRecord()))
    assert entry.parent.as_posix() == "/"
    assert entry.path.as_posix() == "/boot"
    assert entry.type == "directory"

    entry = records[1337]
    assert isinstance(entry, type(MLocateRecord()))
    assert entry.parent.as_posix() == "/usr/lib/x86_64-linux-gnu/perl-base/unicore/lib/Bc"
    assert entry.path.as_posix() == "/usr/lib/x86_64-linux-gnu/perl-base/unicore/lib/Bc/WS.pl"
    assert entry.type == "file"
    assert entry.ts == datetime(2024, 1, 15, 12, 16, 7, tzinfo=timezone.utc)



from __future__ import annotations

import re
from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import export
from dissect.target.plugins.apps.remoteaccess.remoteaccess import (
    GENERIC_FILE_TRANSFER_RECORD_FIELDS,
    GENERIC_LOG_RECORD_FIELDS,
    RemoteAccessPlugin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.plugins.general.users import UserDetails
    from dissect.target.target import Target


class AnydeskPlugin(RemoteAccessPlugin):
    """Anydesk plugin."""

    __namespace__ = "anydesk"

    # Anydesk logs when installed as a service
    SERVICE_GLOBS = (
        # Standard client >= Windows 7
        "sysvol/ProgramData/AnyDesk/*.trace",
        # Custom client >= Windows 7
        "sysvol/ProgramData/AnyDesk/ad_*/*.trace",
        # Windows XP / 2003
        "sysvol/Documents and Settings/Public/AnyDesk/*.trace",
        "sysvol/Documents and Settings/Public/AnyDesk/ad_*/*.trace",
        # Standard/Custom client Linux/MacOS
        "var/log/anydesk*/*.trace",
    )

    # Anydesk Filetransfer logs when installed as a service
    FILETRANSFER_SERVICE_LOGS = (
        # File transfer service log
        "sysvol/ProgramData/AnyDesk/file_transfer_trace.txt",
    )

    # User specific Anydesk filetransfer log
    FILETRANSFER_USER_LOGS = (
        # File transfer log
        "AppData/Roaming/AnyDesk/file_transfer_trace.txt",
    )

    # User specific Anydesk logs
    USER_GLOBS = (
        # Standard client Windows
        "AppData/Roaming/AnyDesk/*.trace",
        # Custom client Windows
        "AppData/Roaming/AnyDesk/ad_*/*.trace",
        # Windows XP / 2003
        "AppData/AnyDesk/*.trace",
        # Standard client Linux/MacOS
        ".anydesk/*.trace",
        # Custom client Linux/MacOS
        ".anydesk_ad_*/*.trace",
    )

    RemoteAccessLogRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "remoteaccess/anydesk/log", GENERIC_LOG_RECORD_FIELDS
    )

    RemoteAccessFileTransferRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "remoteaccess/anydesk/filetransfer", GENERIC_FILE_TRANSFER_RECORD_FIELDS
    )

    def __init__(self, target: Target):
        super().__init__(target)

        self.trace_files: set[tuple[TargetPath, UserDetails | None]] = set()
        self.filetransfer_files: set[tuple[TargetPath, UserDetails | None]] = set()

        # Anydesk trace file service globs
        for trace_glob in self.SERVICE_GLOBS:
            for path in self.target.fs.path().glob(trace_glob):
                self.trace_files.add((path, None))

        # Anydesk filetransfer service globs
        for trace_glob in self.FILETRANSFER_SERVICE_LOGS:
            for path in self.target.fs.path().glob(trace_glob):
                self.filetransfer_files.add((path, None))

        for user_details in self.target.user_details.all_with_home():
            # Anydesk trace file user globs
            for trace_glob in self.USER_GLOBS:
                for path in user_details.home_path.glob(trace_glob):
                    self.trace_files.add((path, user_details.user))

            # Anydesk filetransfer User globs
            for trace_glob in self.FILETRANSFER_USER_LOGS:
                for path in user_details.home_path.glob(trace_glob):
                    self.filetransfer_files.add((path, user_details.user))

    def check_compatible(self) -> None:
        if not self.trace_files and not self.filetransfer_files:
            raise UnsupportedPluginError("No Anydesk files found on target")

    @export(record=RemoteAccessLogRecord)
    def logs(self) -> Iterator[RemoteAccessLogRecord]:
        """Parse AnyDesk trace files.

        AnyDesk is a remote desktop application and can be used by adversaries to get (persistent) access to a machine.
        Log files (.trace files) can be stored on various locations, based on target OS and client type.
        Timestamps in trace files do not carry a time zone designator (TZD) but are in fact UTC.

        References:
            - https://www.inversecos.com/2021/02/forensic-analysis-of-anydesk-logs.html
            - https://support.anydesk.com/knowledge/trace-files#trace-file-locations
        """
        for path, user in self.trace_files:
            for line in path.open("rt", errors="backslashreplace"):
                if not (line := line.strip()) or "* * * * * * * * * * * * * *" in line:
                    continue

                try:
                    level, ts_date, ts_time, message = line.split(" ", 3)

                    timestamp = datetime.strptime(f"{ts_date} {ts_time}", "%Y-%m-%d %H:%M:%S.%f").replace(
                        tzinfo=timezone.utc
                    )
                    message = re.sub(r"\s\s+", " ", f"{level} {message}")

                    yield self.RemoteAccessLogRecord(
                        ts=timestamp,
                        message=message,
                        source=path,
                        _target=self.target,
                        _user=user,
                    )
                except ValueError as e:
                    self.target.log.warning("Could not parse log line in file %s: '%s'", path, line)
                    self.target.log.debug("", exc_info=e)

    @export(record=RemoteAccessFileTransferRecord)
    def filetransfer(self) -> Iterator[RemoteAccessFileTransferRecord]:
        """Parse AnyDesk filetransfer files.

        AnyDesk is a remote desktop application and can be used by adversaries to get (persistent) access to a machine.
        File transfer (``file_transfer_trace.txt``) files show what files are downloaded to a system.
        Timestamps in trace files do not carry a time zone designator (TZD) but are in fact UTC.
        """
        for path, user in self.filetransfer_files:
            for line in path.open("rt", encoding="utf-16-le", errors="backslashreplace"):
                if not (line := line.strip()):
                    continue

                try:
                    # Example log entry:
                    # Clipboard\t\t2025-01-01, 13:37\tstart \tdownload\t'malware.exe' (~0 B out of 4.20 MiB)
                    timestamp = line.split("\t")[2]
                    timestamp = datetime.strptime(timestamp, "%Y-%m-%d, %H:%M").replace(tzinfo=timezone.utc)
                    # Replace log lines tabs with spaces for readability
                    message = line.replace("\t", " ")

                    # Attempt to extract 'filename' from message.
                    filename = match.group(1) if (match := re.search(r"\'([^\']*)\'", message)) else None

                    yield self.RemoteAccessFileTransferRecord(
                        ts=timestamp,
                        message=message,
                        filename=filename,
                        source=path,
                        _target=self.target,
                        _user=user,
                    )
                except ValueError as e:
                    self.target.log.warning("Could not parse log line in file %s: '%s'", path, line)
                    self.target.log.debug("", exc_info=e)



from __future__ import annotations

import textwrap
from datetime import datetime, timezone
from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.apps.container.podman import PodmanPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.fixture
def target_unix_podman(target_unix_users: Target, fs_unix: VirtualFilesystem) -> Target:
    for id, file in [
        ("ae30bde8949f4d4e5b90ad839bbbaffa03db7d9eccbcad7163c34665084d1b70", "httpd"),
        ("4e82f2c6d0ba1a41eacaa5622fcbb9c4e22c9531e6345291a68f6a2219ac9d1a", "nginx"),
        ("bb44c71b8e6c00ba0d4bd483ede69ec32930ca8a30abee4a7f2aadb39cee4988", "debian"),
    ]:
        fs_unix.map_file(
            f"/home/user/.local/share/containers/storage/overlay-containers/{id}/userdata/config.json",
            absolute_path(f"_data/plugins/apps/container/podman/config.json-{file}"),
        )

    fs_unix.map_file(
        "/home/user/.local/share/containers/storage/db.sql",
        absolute_path("_data/plugins/apps/container/podman/db.sql"),
    )

    fs_unix.map_file(
        "/home/user/.local/share/containers/storage/overlay-images/images.json",
        absolute_path("_data/plugins/apps/container/podman/images.json"),
    )

    return target_unix_users


def test_podman_images(target_unix_podman: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we can detect Podman images on a target based on an ``images.json`` file.

    Structure of a Podman OCI image on disk::

        $PODMAN/storage/overlay-images/
                                      /images.json
                                      /images.lock
                                      /<HASH>/
                                             /=<BASE64>
                                             /manifest
    """

    target_unix_podman.add_plugin(PodmanPlugin)
    records = list(target_unix_podman.container.images())

    assert sorted([f"{r.name}:{r.tag}" for r in records]) == [
        "docker.io/library/alpine:latest",
        "docker.io/library/debian:latest",
        "docker.io/library/nginx:latest",
        "docker.io/library/ubuntu:latest",
    ]

    assert records[0].name == "docker.io/library/nginx"
    assert records[0].tag == "latest"
    assert records[0].image_id == "4cad75abc83d"
    assert records[0].hash == "4cad75abc83d5ca6ee22053d85850676eaef657ee9d723d7bef61179e1e1e485"
    assert records[0].created == datetime(2025, 2, 5, 21, 27, 16, tzinfo=timezone.utc)
    assert records[0].source == "/home/user/.local/share/containers/storage/overlay-images/images.json"


def test_podman_containers(target_unix_podman: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we can detect Podman containers on a target based on a SQLite3 database."""

    target_unix_podman.add_plugin(PodmanPlugin)
    records = list(target_unix_podman.container.containers())

    assert sorted([r.name for r in records]) == [
        "boring_mirzakhani",
        "fervent_proskuriakova",
        "hardcore_khayyam",
        "zen_taussig",
    ]

    assert records[0].container_id == "bb44c71b8e6c00ba0d4bd483ede69ec32930ca8a30abee4a7f2aadb39cee4988"
    assert records[0].image == "docker.io/library/debian:latest"
    assert records[0].image_id == "1fd9a3236e02e50084b18aff689d466641759f4e9e5fee930e194a605081be65"
    assert records[0].command == "bash"
    assert records[0].created == datetime(2025, 4, 9, 11, 37, 41, 694673, tzinfo=timezone.utc)
    assert records[0].running
    assert records[0].pid == 58526
    assert records[0].started == datetime(2025, 4, 9, 11, 37, 42, 68128, tzinfo=timezone.utc)
    assert records[0].finished == datetime(1, 1, 1, tzinfo=timezone.utc)
    assert records[0].ports == []
    assert records[0].name == "hardcore_khayyam"
    assert records[0].volumes == ["/tmp/host-folder:/data"]
    assert records[0].environment == [
        "PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin",
        "container=podman",
        "TERM=xterm",
    ]
    assert (
        records[0].mount_path
        == "/home/user/.local/share/containers/storage/overlay/f351129587e2bb1da9ba4f03dcd22e1c838cd4f20dcc70e6da72381d2905b913"  # noqa: E501
    )
    assert (
        records[0].config_path
        == "/home/user/.local/share/containers/storage/overlay-containers/bb44c71b8e6c00ba0d4bd483ede69ec32930ca8a30abee4a7f2aadb39cee4988/userdata/config.json"  # noqa: E501
    )
    assert (
        records[0].image_path
        == "/home/user/.local/share/containers/storage/overlay-images/1fd9a3236e02e50084b18aff689d466641759f4e9e5fee930e194a605081be65"  # noqa: E501
    )
    assert records[0].source == "/home/user/.local/share/containers/storage/db.sql"

    assert records[-1].image == "docker.io/library/nginx:latest"
    assert records[-1].command == "nginx -g daemon off;"
    assert records[-1].volumes == ["/tmp/host-folder/host-file.txt:/data/container-file.txt"]
    assert records[-1].ports == ["0.0.0.0:8080->80/tcp"]


def test_podman_logs(target_unix_podman: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we can parse non-default Podman json-file log driver entries."""

    fs_unix.map_file(
        "/home/user/.local/share/containers/storage/overlay-containers/350f9f5489aebd1d33aa9ed17450270f5a5502d8548df8085d88eccccf2182f1/userdata/ctr.log",
        absolute_path("_data/plugins/apps/container/podman/ctr.log"),
    )

    target_unix_podman.add_plugin(PodmanPlugin)
    records = list(target_unix_podman.container.logs())

    assert len(records) == 51

    assert records[0].ts == datetime(2025, 7, 17, 13, 59, 55, 516060, tzinfo=timezone.utc)
    assert records[0].container == "350f9f5489aebd1d33aa9ed17450270f5a5502d8548df8085d88eccccf2182f1"
    assert records[0].stream == "stdout"
    assert records[0].message == "\x1b[?2004h\x1b]0;root@350f9f5489ae: /\x07root@350f9f5489ae:/# id"
    assert (
        records[0].source
        == "/home/user/.local/share/containers/storage/overlay-containers/350f9f5489aebd1d33aa9ed17450270f5a5502d8548df8085d88eccccf2182f1/userdata/ctr.log"  # noqa: E501
    )

    assert "\n".join([r.message for r in records]) == textwrap.dedent("""\
    \x1b[?2004h\x1b]0;root@350f9f5489ae: /\x07root@350f9f5489ae:/# id
    uid=0(root) gid=0(root) groups=0(root)
    \x1b[?2004l\x1b[?2004h\x1b]0;root@350f9f5489ae: /\x07root@350f9f5489ae:/# whoami
    root
    \x1b[?2004l\x1b[?2004h\x1b]0;root@350f9f5489ae: /\x07root@350f9f5489ae:/# ls -lah .
    total 56K
    dr-xr-xr-x   1 root   root    4.0K Jul 17 13:59 \x1b[0m\x1b[01;34m.\x1b[0m
    dr-xr-xr-x   1 root   root    4.0K Jul 17 13:59 \x1b[01;34m..\x1b[0m
    lrwxrwxrwx   1 root   root       7 Apr 22  2024 \x1b[01;36mbin\x1b[0m -> \x1b[01;34musr/bin\x1b[0m
    drwxr-xr-x   2 root   root    4.0K Apr 22  2024 \x1b[01;34mboot\x1b[0m
    drwxr-xr-x   5 root   root     360 Jul 17 13:59 \x1b[01;34mdev\x1b[0m
    drwxr-xr-x   1 root   root    4.0K Jul 17 13:59 \x1b[01;34metc\x1b[0m
    drwxr-xr-x   3 root   root    4.0K Jul 14 14:14 \x1b[01;34mhome\x1b[0m
    lrwxrwxrwx   1 root   root       7 Apr 22  2024 \x1b[01;36mlib\x1b[0m -> \x1b[01;34musr/lib\x1b[0m
    lrwxrwxrwx   1 root   root       9 Apr 22  2024 \x1b[01;36mlib64\x1b[0m -> \x1b[01;34musr/lib64\x1b[0m
    drwxr-xr-x   2 root   root    4.0K Jul 14 14:08 \x1b[01;34mmedia\x1b[0m
    drwxr-xr-x   2 root   root    4.0K Jul 14 14:08 \x1b[01;34mmnt\x1b[0m
    drwxr-xr-x   2 root   root    4.0K Jul 14 14:08 \x1b[01;34mopt\x1b[0m
    dr-xr-xr-x 278 nobody nogroup    0 Jul 17 13:59 \x1b[01;34mproc\x1b[0m
    drwx------   2 root   root    4.0K Jul 14 14:14 \x1b[01;34mroot\x1b[0m
    drwxr-xr-x   1 root   root    4.0K Jul 17 13:59 \x1b[01;34mrun\x1b[0m
    lrwxrwxrwx   1 root   root       8 Apr 22  2024 \x1b[01;36msbin\x1b[0m -> \x1b[01;34musr/sbin\x1b[0m
    drwxr-xr-x   2 root   root    4.0K Jul 14 14:08 \x1b[01;34msrv\x1b[0m
    dr-xr-xr-x  13 nobody nogroup    0 Jul 17 13:59 \x1b[01;34msys\x1b[0m
    drwxrwxrwt   2 root   root    4.0K Jul 14 14:14 \x1b[30;42mtmp\x1b[0m
    drwxr-xr-x  12 root   root    4.0K Jul 14 14:08 \x1b[01;34musr\x1b[0m
    drwxr-xr-x  11 root   root    4.0K Jul 14 14:14 \x1b[01;34mvar\x1b[0m
    \x1b[?2004l\x1b[?2004h\x1b]0;root@350f9f5489ae: /\x07root@350f9f5489ae:/# cat /etc/passwd
    root:x:0:0:root:/root:/bin/bash
    daemon:x:1:1:daemon:/usr/sbin:/usr/sbin/nologin
    bin:x:2:2:bin:/bin:/usr/sbin/nologin
    sys:x:3:3:sys:/dev:/usr/sbin/nologin
    sync:x:4:65534:sync:/bin:/bin/sync
    games:x:5:60:games:/usr/games:/usr/sbin/nologin
    man:x:6:12:man:/var/cache/man:/usr/sbin/nologin
    lp:x:7:7:lp:/var/spool/lpd:/usr/sbin/nologin
    mail:x:8:8:mail:/var/mail:/usr/sbin/nologin
    news:x:9:9:news:/var/spool/news:/usr/sbin/nologin
    uucp:x:10:10:uucp:/var/spool/uucp:/usr/sbin/nologin
    proxy:x:13:13:proxy:/bin:/usr/sbin/nologin
    www-data:x:33:33:www-data:/var/www:/usr/sbin/nologin
    backup:x:34:34:backup:/var/backups:/usr/sbin/nologin
    list:x:38:38:Mailing List Manager:/var/list:/usr/sbin/nologin
    irc:x:39:39:ircd:/run/ircd:/usr/sbin/nologin
    _apt:x:42:65534::/nonexistent:/usr/sbin/nologin
    nobody:x:65534:65534:nobody:/nonexistent:/usr/sbin/nologin
    ubuntu:x:1000:1000:Ubuntu:/home/ubuntu:/bin/bash
    \x1b[?2004l\x1b[?2004h\x1b]0;root@350f9f5489ae: /\x07root@350f9f5489ae:/# echo "hidden\x08 \x08\x08 \x08\x08 \x08\x08 \x08\x08 \x08\x08 \x08hello!"
    \x1b[?2004l
    \x1b[?2004l
    exit""")  # noqa: E501



from __future__ import annotations

from typing import TYPE_CHECKING

from flow.record import GroupedRecord

from dissect.target.exceptions import InvalidTaskError, UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export
from dissect.target.plugins.os.windows.tasks.job import AtTask
from dissect.target.plugins.os.windows.tasks.records import (
    BaseTriggerRecord,
    BootTriggerRecord,
    CalendarTriggerRecord,
    ComHandlerRecord,
    DailyTriggerRecord,
    EventTriggerRecord,
    ExecRecord,
    IdleTriggerRecord,
    LogonTriggerRecord,
    MonthlyDateTriggerRecord,
    MonthlyDowTriggerRecord,
    RegistrationTrigger,
    SendEmailRecord,
    SessionStateChangeTriggerRecord,
    ShowMessageRecord,
    TimeTriggerRecord,
    WeeklyTriggerRecord,
    WnfTriggerRecord,
)
from dissect.target.plugins.os.windows.tasks.xml import ScheduledTasks

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

TaskRecord = TargetRecordDescriptor(
    "filesystem/windows/task",
    [
        ("path", "task_path"),
        ("string", "uri"),
        ("string", "security_descriptor"),
        ("string", "source"),
        ("datetime", "date"),
        ("datetime", "last_run_date"),
        ("string", "author"),
        ("string", "version"),
        ("string", "description"),
        ("string", "documentation"),
        ("string", "task_name"),
        ("string", "app_name"),
        ("string", "args"),
        ("string", "start_in"),
        ("string", "comment"),
        ("string", "run_as"),
        ("string", "cpassword"),
        ("boolean", "enabled"),
        ("string", "action"),
        ("string", "principal_id"),
        ("string", "user_id"),
        ("string", "logon_type"),
        ("string", "group_id"),
        ("string", "display_name"),
        ("string", "run_level"),
        ("string", "process_token_sid_type"),
        ("string", "required_privileges"),
        ("string", "restart_on_failure_interval"),
        ("string", "restart_on_failure_count"),
        ("string", "mutiple_instances_policy"),
        ("boolean", "disallow_start_on_batteries"),
        ("boolean", "stop_going_on_batteries"),
        ("boolean", "allow_start_on_demand"),
        ("boolean", "start_when_available"),
        ("string", "network_profile_name"),
        ("boolean", "run_only_network_available"),
        ("boolean", "wake_to_run"),
        ("boolean", "hidden"),
        ("string", "delete_expired_task_after"),
        ("string", "idle_duration"),
        ("string", "idle_wait_timeout"),
        ("boolean", "idle_stop_on_idle_end"),
        ("boolean", "idle_restart_on_idle"),
        ("string", "network_settings_name"),
        ("string", "network_settings_id"),
        ("string", "execution_time_limit"),
        ("string", "priority"),
        ("boolean", "run_only_idle"),
        ("boolean", "unified_scheduling_engine"),
        ("boolean", "disallow_start_on_remote_app_session"),
        ("string", "data"),
        ("string", "raw_data"),
    ],
)

TriggerRecord = TargetRecordDescriptor(
    "filesystem/windows/task/trigger",
    {
        ("string", "uri"),
        *BaseTriggerRecord.target_fields,
        *BootTriggerRecord.target_fields,
        *CalendarTriggerRecord.target_fields,
        *DailyTriggerRecord.target_fields,
        *EventTriggerRecord.target_fields,
        *IdleTriggerRecord.target_fields,
        *LogonTriggerRecord.target_fields,
        *MonthlyDateTriggerRecord.target_fields,
        *MonthlyDowTriggerRecord.target_fields,
        *RegistrationTrigger.target_fields,
        *SessionStateChangeTriggerRecord.target_fields,
        *TimeTriggerRecord.target_fields,
        *WeeklyTriggerRecord.target_fields,
        *WnfTriggerRecord.target_fields,
    },
)

ActionRecord = TargetRecordDescriptor(
    "filesystem/windows/task/action",
    {
        ("string", "uri"),
        *ComHandlerRecord.target_fields,
        *ExecRecord.target_fields,
        *SendEmailRecord.target_fields,
        *ShowMessageRecord.target_fields,
    },
)


class TasksPlugin(Plugin):
    """Plugin for retrieving scheduled tasks on a Windows system.

    Args:
        target: The target system.
    """

    PATHS = (
        "%windir%/system32/tasks",
        "%windir%/system32/tasks_migrated",
        "%windir%/syswow64/tasks",
        "%windir%/tasks",  # at.exe job file location
    )
    GLOB_PATHS = (
        "%windir%/system32/GroupPolicy/DataStore/*/Machine/Preferences/ScheduledTasks/*",
        "sysvol/ProgramData/Microsoft/*/Preferences/ScheduledTasks/*",
    )

    def __init__(self, target: Target):
        super().__init__(target)
        self.task_files = []

        for path in self.GLOB_PATHS:
            start_path, pattern = path.split("*", 1)
            for entry in self.target.resolve(start_path).rglob("*" + pattern):
                if entry.is_file() and entry.suffix == ".xml":
                    self.task_files.append(entry)

        for file_path in self.PATHS:
            fpath = self.target.resolve(file_path)
            if not fpath.exists():
                continue

            for entry in fpath.rglob("*"):
                if entry.is_file() and (entry.suffix.lower() == ".job" or not entry.suffix):
                    self.task_files.append(entry)

    def check_compatible(self) -> None:
        if len(self.task_files) == 0:
            raise UnsupportedPluginError("No task files")

    @export(record=[TaskRecord, TriggerRecord, ActionRecord])
    @arg(
        "--group",
        action="store_true",
        help="group each trigger and action record together with its corresponding parent task record",
    )
    def tasks(self, group: bool = False) -> Iterator[TaskRecord | TriggerRecord | ActionRecord | GroupedRecord]:
        """Return all scheduled tasks on a Windows system.

        On a Windows system, a scheduled task is a program or script that is executed on a specific time or at specific
        intervals. An adversary may leverage such scheduled tasks to gain persistence on a system.

        References:
            - https://en.wikipedia.org/wiki/Windows_Task_Scheduler

        Yields:
            The scheduled tasks found on the target.
        """
        target_tz = self.target.datetime.tzinfo

        for task_file in self.task_files:
            if not task_file.suffix or task_file.suffix == ".xml":
                try:
                    task_objects = ScheduledTasks(task_file).tasks
                except InvalidTaskError as e:
                    self.target.log.warning("Invalid task file encountered: %s", task_file)
                    self.target.log.debug("", exc_info=e)
                    continue
            else:
                task_objects = [AtTask(task_file, target_tz)]

            for task_object in task_objects:
                record_kwargs = {}
                for attr in TaskRecord.fields:
                    record_kwargs[attr] = getattr(task_object, attr, None)

                task_record = TaskRecord(**record_kwargs, _target=self.target)
                yield task_record

                # Actions
                for action in task_object.get_actions():
                    if group:
                        record = GroupedRecord("filesystem/windows/task/grouped", [task_record, action])
                    else:
                        record = ActionRecord(**action._asdict(), uri=task_record.uri, _target=self.target)
                    yield record

                # Triggers
                for trigger in task_object.get_triggers():
                    if group:
                        record = GroupedRecord("filesystem/windows/task/grouped", [task_record, trigger])
                    else:
                        record = TriggerRecord(**trigger._asdict(), uri=task_record.uri, _target=self.target)
                    yield record



from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING, BinaryIO

from dissect.fve import bde

from dissect.target.exceptions import VolumeSystemError
from dissect.target.helpers.keychain import KeyType
from dissect.target.helpers.logging import get_logger
from dissect.target.volume import EncryptedVolumeSystem, Volume

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.util.stream import AlignedStream


log = get_logger(__name__)


class BitlockerVolumeSystemError(VolumeSystemError):
    pass


class BitlockerVolumeSystem(EncryptedVolumeSystem):
    __type__ = "bitlocker"

    def __init__(self, fh: BinaryIO | list[BinaryIO], *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.bde = bde.BDE(fh)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        return bde.is_bde_volume(fh)

    def _volumes(self) -> Iterator[Volume]:
        if isinstance(self.fh, Volume):
            volume_details = {
                "number": self.fh.number,
                "offset": self.fh.offset,
                "vtype": self.fh.type,
                "name": self.fh.name,
                "guid": self.fh.guid,
            }
        else:
            volume_details = {
                "number": 1,
                "offset": 0,
                "vtype": None,
                "name": None,
            }

        stream = self.unlock_volume()
        yield Volume(
            fh=stream,
            size=stream.size,
            raw=self.fh,
            disk=self.disk,
            vs=self,
            **volume_details,
        )

    def unlock_with_passphrase(self, passphrase: str, is_wildcard: bool = False) -> None:
        try:
            self.bde.unlock_with_passphrase(passphrase)
            log.debug("Unlocked BDE volume with provided passphrase")
        except ValueError:
            if not is_wildcard:
                log.exception("Failed to unlock BDE volume with provided passphrase")

    def unlock_with_recovery_key(self, recovery_key: str, is_wildcard: bool = False) -> None:
        try:
            self.bde.unlock_with_recovery_password(recovery_key)
            log.debug("Unlocked BDE volume with recovery key")
        except ValueError:
            if not is_wildcard:
                log.exception("Failed to unlock BDE volume with recovery password")

    def unlock_with_bek_file(self, bek_file: pathlib.Path, is_wildcard: bool = False) -> None:
        if not bek_file.exists():
            if not is_wildcard:
                log.error("Provided BEK file does not exist: %s", bek_file)
            return

        with bek_file.open(mode="rb") as fh:
            try:
                self.bde.unlock_with_bek(fh)
                log.debug("Unlocked BDE volume with BEK file %s", bek_file)
            except ValueError:
                if not is_wildcard:
                    log.exception("Failed to unlock BDE volume with BEK file %s", bek_file)

    def unlock_with_fvek(self, raw_key: bytes, is_wildcard: bool = False) -> None:
        try:
            self.bde.unlock_with_fvek(raw_key)
        except ValueError:
            if not is_wildcard:
                log.exception("Failed to unlock BDE volume with raw FVEK key (%r)", raw_key)

    def unlock_volume(self) -> AlignedStream:
        if self.bde.has_clear_key():
            self.bde.unlock_with_clear_key()
        else:
            identifiers = [str(i) for i in self.bde.identifiers]
            keys = self.get_keys_for_identifiers(identifiers) + self.get_keys_without_identifier()

            for key in keys:
                if key.key_type == KeyType.PASSPHRASE and self.bde.has_passphrase():
                    self.unlock_with_passphrase(key.value, key.is_wildcard)
                elif key.key_type == KeyType.RECOVERY_KEY and self.bde.has_recovery_password():
                    self.unlock_with_recovery_key(key.value, key.is_wildcard)
                elif key.key_type == KeyType.FILE:
                    bek_file = pathlib.Path(key.value)
                    self.unlock_with_bek_file(bek_file, key.is_wildcard)
                elif key.key_type == KeyType.RAW:
                    self.unlock_with_fvek(key.value, key.is_wildcard)

                if self.bde.unlocked:
                    log.info("Volume %s with identifiers %s unlocked with %s", self.fh, identifiers, key)
                    break

        if self.bde.unlocked:
            return self.bde.open()
        raise BitlockerVolumeSystemError("Failed to unlock BDE volume")



from __future__ import annotations

import itertools
from enum import Enum
from functools import lru_cache
from typing import TYPE_CHECKING, Final, Literal, NamedTuple

from defusedxml import ElementTree
from flow.record.fieldtypes import windows_path

from dissect.target.exceptions import RegistryError, UnsupportedPluginError
from dissect.target.helpers import fsutil
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.helpers.utils import to_list
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path
    from xml.etree.ElementTree import Element

    from dissect.target.helpers.regutil import KeyCollection
    from dissect.target.target import Target

OfficeStartupItem = TargetRecordDescriptor(
    "application/productivity/msoffice/startup_item",
    [("path", "path"), ("datetime", "creation_time"), ("datetime", "modification_time")],
)

# Web add-in
OfficeWebAddinRecord = TargetRecordDescriptor(
    "application/productivity/msoffice/web_addin",
    [
        ("path", "manifest"),
        ("datetime", "modification_time"),
        ("string", "name"),
        ("string", "version"),
        ("string", "provider_name"),
        ("string[]", "source_locations"),
    ],
)

# COM and VSTO add-ins
OfficeNativeAddinRecord = TargetRecordDescriptor(
    "application/productivity/msoffice/native_addin",
    [
        ("path", "manifest"),
        ("datetime", "modification_time"),
        ("string", "name"),
        ("string", "type"),
        ("path[]", "codebases"),
        ("boolean", "loaded"),
        ("string", "load_behavior"),
    ],
)


class ClickOnceDeploymentManifestParser:
    """Parser to extact information out ClickOnce deployment manifest files.

    Currently only extracts codebase information. Also handles nested manifests.
    Can be extended to a .NET assembly parser in the future.
    """

    XML_NAMESPACE: Final[dict[str, str]] = {"": "urn:schemas-microsoft-com:asm.v2"}

    class Assembly(NamedTuple):
        installed: bool
        codebase: Path

    def __init__(self, root_manifest_path: Path, target: Target, user_sid: str):
        self.root_manifest_path = root_manifest_path
        self._target = target
        self._user_sid = user_sid
        self._visited_manifests: set[Path] = set()
        self._codebases: set[Path] = set()

    def find_codebases(self, manifest_path: str) -> set[Path]:
        """Dig for executables given a manifest."""

        assemblies = self._parse_manifest(manifest_path)
        # Ignore pre-installed assemblies
        return {assembly.codebase for assembly in assemblies if assembly.installed}

    def _parse_manifest(self, manifest_path: Path) -> set[Assembly]:
        # See https://learn.microsoft.com/en-us/visualstudio/deployment/clickonce-deployment-manifest?view=vs-2022

        if manifest_path in self._visited_manifests:
            return self._codebases  # Prevent cycles

        self._visited_manifests.add(manifest_path)
        try:
            manifest_tree: Element = ElementTree.fromstring(manifest_path.read_text("utf-8-sig"))
        except Exception as e:
            self._target.log.warning("Error parsing manifest %s", manifest_path)
            self._target.log.debug("", exc_info=e)
            return set()

        dependent_assemblies: set[ClickOnceDeploymentManifestParser.Assembly] = set()
        dependent_assembly_elements = manifest_tree.findall(".//dependentAssembly", self.XML_NAMESPACE)
        for dependent_assembly_element in dependent_assembly_elements:
            dependent_assemblies |= self._parse_dependent_assembly(dependent_assembly_element, manifest_path.parent)

        return dependent_assemblies

    def _parse_dependent_assembly(self, dependent_assembly: Element, cwd: Path) -> set[Assembly]:
        # See https://learn.microsoft.com/en-us/visualstudio/deployment/dependency-element-clickonce-deployment?view=vs-2022#dependentassembly # noqa: E501

        if not (codebase_str_path := dependent_assembly.get("codebase")):
            return set()

        codebase_str_path = fsutil.abspath(codebase_str_path, str(cwd), alt_separator=self._target.fs.alt_separator)
        codebase_path: Path = self._target.fs.path(codebase_str_path)
        if not codebase_path.exists():
            return set()  # Ignore files which are not actually installed, for example due to language settings

        installed = dependent_assembly.get("dependencyType") == "install"
        if codebase_path.name.endswith(".manifest") and installed:
            return self._parse_manifest(codebase_path)  # Yes, a codebase can point to another manifest

        return {self.Assembly(installed, codebase_path)}


class LoadBehavior(Enum):
    """Values that specify the run time behavior of the VSTO add-in."""

    Manual = 1
    Autostart = 2
    OnDemand = 3
    FistTime = 4


class NativePluginStatus(NamedTuple):
    loaded: bool
    load_behavior: LoadBehavior


class MSOffice(Plugin):
    """Microsoft Office productivity suite plugin."""

    __namespace__ = "msoffice"

    HIVES = ("HKLM", "HKCU")
    OFFICE_KEY = "Software\\Microsoft\\Office"
    OFFICE_COMPONENTS = ("Access", "Excel", "Outlook", "PowerPoint", "Word", "OneNote")
    ADD_IN_KEY = "Addins"
    OFFICE_DEFAULT_USER_STARTUP = (
        "%APPDATA%/Microsoft/Templates",
        "%APPDATA%/Microsoft/Word/Startup",
        "%APPDATA%/Microsoft/Excel/XLSTART",
        "%APPDATA%/Microsoft/Outlook/Startup",
        "%APPDATA%/Microsoft/PowerPoint/Startup",
    )

    OFFICE_DEFAULT_ROOT = "C:/Program Files/Microsoft Office/root/Office16/"

    # Office is fixed at version 16.0 since Microsoft Office 2016 (released in 2015)
    # Powerpoint and Outlook do not have a alternate startup folder
    OFFICE_STARTUP_OPTIONS = (
        ("Software\\Microsoft\\Office\\16.0\\Word\\Options", "STARTUP-PATH"),
        ("Software\\Microsoft\\Office\\16.0\\Word\\Options", "UserTemplates"),
        ("Software\\Microsoft\\Office\\16.0\\Excel\\Options", "AltStartup"),
    )

    CLASSES_ROOTS = (
        "HKCR",
        # Click To Run Application Virtualization:
        "HKLM\\SOFTWARE\\Microsoft\\Office\\ClickToRun\\REGISTRY\\MACHINE\\Software\\Classes",
        # For 32-bit software running under 64-bit Windows:
        "HKLM\\SOFTWARE\\Wow6432Node\\Classes",
    )

    def __init__(self, target: Target):
        super().__init__(target)
        self._office_install_root = lru_cache(32)(self._office_install_root)

    def check_compatible(self) -> None:
        if not self.target.has_function("registry") or not list(self.target.registry.keys(f"HKLM\\{self.OFFICE_KEY}")):
            raise UnsupportedPluginError("Registry key not found: %s", self.OFFICE_KEY)

    @export(record=OfficeWebAddinRecord)
    def web(self) -> Iterator[OfficeWebAddinRecord]:
        """Returns all available Web add-ins cached in the WEF (Web Extension Framework) folder.

        Office Web Add-ins are web-based applications that extend the functionality of Office applications like Word, Excel, and Outlook.
        These add-ins can interact with the content in Office documents and provide additional features and capabilities.
        The WEF folder contains cached data and manifests for Office Web Add-ins.
        The manifest includes information about the add-ins, such as their source locations, display names, and other metadata.

        References:
            - https://learn.microsoft.com/en-us/office/dev/add-ins/overview/office-add-ins

        Yields a ``OfficeWebAddinRecord`` with fields:

        .. code-block:: text

            manifest (path): The full path to the manifest in the WEF folder.
            modification_time (datetime): The modification time of the manifest.
            name (string): The display name of the add-in.
            version (string): The version of the add-in.
            provider_name (string): The provider name of the add-in.
            source_locations (string[]): URLs referencing the web assets of the add-in (such as javascript and html files).
        """  # noqa: E501

        for manifest_file in self._wef_cache_folders():
            try:
                yield self._parse_web_addin_manifest(manifest_file)
            except Exception as e:  # noqa: PERF203
                self.target.log.warning("Error parsing web-addin manifest %s", manifest_file)
                self.target.log.debug("", exc_info=e)

    @export(record=OfficeNativeAddinRecord)
    def native(self) -> Iterator[OfficeNativeAddinRecord]:
        """Returns all native (COM / VSTO) add-ins by parsing the registry and manifest files.

        COM (Component Object Model) is a binary-interface standard developed by Microsoft that enables software components to communicate with each other.
        COM plugins for Microsoft Office applications, such as Word, Excel, and Outlook, are typically used to extend the functionality of these programs by integrating custom features.
        COM plugins interact directly with Office applications through COM interfaces, offering a low-level approach to automation.

        VSTO is a set of tools provided by Microsoft to create Office add-ins using the .NET Framework.
        VSTO plugins are more modern than COM plugins and leverage managed code. They are typically developed in C# or VB.NET using Visual Studio.

        Both COM and VSTO add-ins are registered in the Windows registry, where they are associated with specific Office applications and configured to load automatically or on demand.

        References:
            - https://learn.microsoft.com/en-us/office/dev/add-ins/overview/office-add-ins
            - https://learn.microsoft.com/en-us/visualstudio/vsto/registry-entries-for-vsto-add-ins

        Yields a ``OfficeNativeAddinRecord`` with fields:

        .. code-block:: text

            manifest (path): The full path to the manifest of a VSTO plugin. ``None`` for COM plugins.
            modification_time (datetime): The modification time of the registry key of the plugin.
            name (string): The name of the add-in.
            type (string): The type of the add-in, either "com" or "vsto".
            codebases (path[]): The full paths to the executables associated with the add-in.
            loaded (boolean): Whether the add-in is currently loaded.
            load_behavior (string): The load behavior of the add-in, e.g., "Autostart", "Manual", "OnDemand", "FirstTime".
        """  # noqa: E501

        addin_path_tuples = itertools.product(self.HIVES, [self.OFFICE_KEY], self.OFFICE_COMPONENTS, [self.ADD_IN_KEY])
        addin_paths = ["\\".join(addin_path_tuple) for addin_path_tuple in addin_path_tuples]
        for addin_key in self.target.registry.keys(addin_paths):
            for addin in addin_key.subkeys():
                key_owner = self.target.registry.get_user(addin)
                sid = key_owner.sid if key_owner else None

                if manifest_path_str := addin.value("Manifest", None).value:
                    addin_type = "vsto"
                    executables = self._parse_vsto_manifest(manifest_path_str, sid)
                else:
                    addin_type = "com"
                    dll_str = self._lookup_com_executable(addin.name)
                    executables = to_list(self.target.resolve(dll_str, sid)) if dll_str else []

                native_plugin_status = self._parse_plugin_status(addin)
                yield OfficeNativeAddinRecord(
                    name=addin.value("FriendlyName", None).value,
                    modification_time=addin.timestamp,
                    loaded=native_plugin_status.loaded if native_plugin_status else None,
                    load_behavior=native_plugin_status.load_behavior.name if native_plugin_status else None,
                    type=addin_type,
                    manifest=windows_path(manifest_path_str) if manifest_path_str else None,
                    codebases=executables,
                )

    @export(record=OfficeStartupItem)
    def startup(self) -> Iterator[OfficeStartupItem]:
        """Returns all startup items found in Microsoft Office startup folders.

        Office startup folders are specific directories where Microsoft Office looks add-ins, macros, templates, or custom scripts.
        These are used to automatically load when the corresponding Office application starts up.
        These folders allow users and administrators to automate launching add-ins, executing scripts, or applying custom settings.

        References:
            - https://pentestlab.blog/2019/12/11/persistence-office-application-startup/

        Yields a ``OfficeStartupItem`` with fields:

        .. code-block:: text

            path (path): The full path to the startup item.
            creation_time (datetime): The creation time of the startup item.
            modification_time (datetime): The modification time of the startup item.
        """  # noqa: E501

        # Get items from default machine-scoped startup folder
        for machine_startup_folder in self._machine_startup_folders():
            yield from self._walk_startup_folder(machine_startup_folder)

        # Get items from default user-scoped startup folder
        for user in self.target.user_details.all_with_home():
            for user_startup_folder in self.OFFICE_DEFAULT_USER_STARTUP:
                yield from self._walk_startup_folder(user_startup_folder, user.user.sid)

        # Get items from alternate machine or user scoped startup folder
        for hive in self.HIVES:
            for options_key, startup_value in self.OFFICE_STARTUP_OPTIONS:
                alt_startup_folders = self.target.registry.values(f"{hive}\\{options_key}", startup_value)
                for alt_startup_folder in alt_startup_folders:
                    user = self.target.registry.get_user(alt_startup_folder)
                    user_sid = user.sid if user else None
                    yield from self._walk_startup_folder(alt_startup_folder.value, user_sid)

    def _wef_cache_folders(self) -> Iterator[Path]:
        """List cache folders which contain office web-addin data."""

        WEB_ADDIN_MANIFEST_GLOB = "AppData/Local/Microsoft/Office/16.0/Wef/**/Manifests/**/*"
        for user_details in self.target.user_details.all_with_home():
            for manifest_path in user_details.home_path.glob(WEB_ADDIN_MANIFEST_GLOB):
                if manifest_path.is_file() and manifest_path.suffix != ".metadata":
                    yield manifest_path

    def _walk_startup_folder(self, startup_folder: str, user_sid: str | None = None) -> Iterator[OfficeStartupItem]:
        """Resolve the given path and return all statup items."""

        resolved_startup_folder: Path = self.target.resolve(startup_folder, user_sid)
        if not resolved_startup_folder.is_dir():
            return

        for current_path, _, plugin_files in resolved_startup_folder.walk():
            for plugin_file in plugin_files:
                item_startup = current_path / plugin_file
                stats = item_startup.stat()
                yield OfficeStartupItem(
                    path=item_startup, creation_time=stats.st_birthtime, modification_time=stats.st_mtime
                )

    def _lookup_com_executable(self, prog_id: str) -> str | None:
        """Lookup the com executable given a prog id using the registry."""

        for classes_root in self.CLASSES_ROOTS:
            try:
                cls_id = self.target.registry.value(f"{classes_root}\\{prog_id}\\CLSID", "(Default)").value
                inproc_key = f"{classes_root}\\CLSID\\{cls_id}\\InprocServer32"
                return self.target.registry.value(inproc_key, "(Default)").value
            except RegistryError:  # noqa: PERF203
                self.target.log.warning("Could not find ProgID %s of COM executable in registry", prog_id)

        return None

    def _parse_vsto_manifest(self, manifest_path_str: str, user_sid: str) -> set[str]:
        """Parse a vsto manifest.

        Non-local manifests, i.e. not ending with suffix "vstolocal" are listed but skipped.
        """

        if not manifest_path_str.endswith("vstolocal"):
            self.target.log.warning("Parsing of remote vsto manifest %s is not supported")
            return set(manifest_path_str)

        manifest_path: Path = self.target.resolve(manifest_path_str.removesuffix("|vstolocal"), user_sid)
        manifest_parser = ClickOnceDeploymentManifestParser(manifest_path, self.target, user_sid)
        return manifest_parser.find_codebases(manifest_path)

    def _parse_web_addin_manifest(self, manifest_path: Path) -> OfficeWebAddinRecord:
        """Parses a web addin manifest."""

        ns = {"": "http://schemas.microsoft.com/office/appforoffice/1.1"}

        manifest_tree: Element = ElementTree.fromstring(manifest_path.read_text("utf-8-sig"))

        source_location_elements = manifest_tree.findall(".//SourceLocation", ns)
        source_locations = [source_location.get("DefaultValue") for source_location in source_location_elements]

        display_name_element = manifest_tree.find(".//DisplayName", ns)
        display_name = display_name_element.get("DefaultValue") if display_name_element is not None else None

        return OfficeWebAddinRecord(
            name=display_name,
            manifest=manifest_path,
            version=manifest_tree.findtext(".//Version", namespaces=ns),
            provider_name=manifest_tree.findtext(".//ProviderName", namespaces=ns),
            source_locations=filter(None, source_locations),
            modification_time=manifest_path.stat().st_mtime,
        )

    def _office_install_root(self, component: Literal["Word", "Excel"]) -> str:
        """Return the installation root for a office component."""

        # Typically, all components share the same root.
        # Curiously enough, the "Common" component has no InstallRoot defined.
        key = f"HKLM\\{self.OFFICE_KEY}\\16.0\\{component}\\InstallRoot"
        try:
            return self.target.registry.value(key, "Path").value
        except RegistryError:
            return self.OFFICE_DEFAULT_ROOT

    def _machine_startup_folders(self) -> Iterator[str]:
        """Return machine-scoped office startup folders."""

        yield fsutil.join(self._office_install_root("Word"), "STARTUP", alt_separator="\\")
        yield fsutil.join(self._office_install_root("Excel"), "XLSTART", alt_separator="\\")
        yield fsutil.join(self._office_install_root("Word"), "Templates", alt_separator="\\")
        yield fsutil.join(self._office_install_root("Word"), "Document Themes", alt_separator="\\")

    def _parse_plugin_status(self, addin: KeyCollection) -> NativePluginStatus | None:
        """Parse the registry value which controls if the add-in autostarts.

        See https://learn.microsoft.com/en-us/visualstudio/vsto/registry-entries-for-vsto-add-ins?view=vs-2022#LoadBehavior
        """

        load_behavior = addin.value("LoadBehavior", None).value

        if load_behavior is None:
            return None
        if load_behavior == 0:
            return NativePluginStatus(loaded=False, load_behavior=LoadBehavior.Manual)
        if load_behavior == 1:
            return NativePluginStatus(loaded=True, load_behavior=LoadBehavior.Manual)
        if load_behavior == 2:
            return NativePluginStatus(loaded=False, load_behavior=LoadBehavior.Autostart)
        if load_behavior == 3:
            return NativePluginStatus(loaded=True, load_behavior=LoadBehavior.Autostart)
        if load_behavior == 8:
            return NativePluginStatus(loaded=False, load_behavior=LoadBehavior.OnDemand)
        if load_behavior == 9:
            return NativePluginStatus(loaded=True, load_behavior=LoadBehavior.OnDemand)
        if load_behavior == 16:
            return NativePluginStatus(loaded=False, load_behavior=LoadBehavior.FistTime)

        return None



from __future__ import annotations

import gzip
import io
from typing import TYPE_CHECKING

import pytest

from dissect.target.containers.fortifw import FortiFirmwareFile, find_xor_key, main

if TYPE_CHECKING:
    from pathlib import Path

# decompressed header of FGT_VM64-v7.4.3.F-build2573-FORTINET.out
FIRMWARE_HEADER = bytes.fromhex(
    """
9ede e4a5 f69f c9aa 92de 92ff b7eb 060a
6475 5c76 7173 6a68 4347 4a4c 774c 7dac
b893 c5e0 81d7 aef0 f58b f3ae d3bf d581
a9f1 a2fb a2fc b0eb 9eca 93c6 83d1 b5fa
9bdb e1a0 f39a d8bb 83cf 83ee a1fd a6c1
e9b1 e2bb e2bc f0ab de8a d386 c391 f5ba
db9b a1e0 b3da 98fb c38f c3ae e1bd e681
a9f1 a2fb a2fc b0eb 9eca 93c6 83d1 b5fa
"""
)


def test_find_xor_key() -> None:
    fh = io.BytesIO(FIRMWARE_HEADER)
    for offset in (0x30, 0x40, 0x31, 0x35):
        fh.seek(offset)
        key = find_xor_key(fh)
        assert key.isalnum()
        assert key == b"aA8BWlDd0EFfCQUh8IAJMKZLmMCNYOzP"

    # incorrect key as at offset 0 it does not decode to zero bytes
    fh.seek(0)
    key = find_xor_key(fh)
    assert not key.isalnum()


@pytest.mark.parametrize(
    ("header", "is_gzipped"),
    [
        pytest.param(FIRMWARE_HEADER, False, id="uncompressed"),
        pytest.param(gzip.compress(FIRMWARE_HEADER), True, id="compressed"),
    ],
)
def test_deobfuscate_firmware_file(header: bytes, is_gzipped: bool) -> None:
    ff = FortiFirmwareFile(io.BytesIO(header))

    # magic bytes
    ff.seek(12)
    assert ff.read(4) == b"\xff\x00\xaa\x55"

    # firmware name
    ff.seek(16)
    assert ff.read(32) == b"FGVM64-7.04-FW-build2573-240201-"

    # test metadata
    assert ff.is_gzipped == is_gzipped
    assert ff.size == len(FIRMWARE_HEADER)


def test_gzip_trailer() -> None:
    trailer_data = b"TRAILER DATA"

    gzip_data = gzip.compress(FIRMWARE_HEADER)
    fh = io.BytesIO(gzip_data + trailer_data)
    ff = FortiFirmwareFile(fh)

    assert ff.is_gzipped
    assert ff.size == len(FIRMWARE_HEADER)
    assert ff.trailer_offset == len(gzip_data)
    assert ff.trailer_data == trailer_data


def test_fortifw_main(tmp_path: Path, capsysbinary: pytest.CaptureFixture[bytes]) -> None:
    fw_path = tmp_path / "fw.bin"
    fw_path.write_bytes(FIRMWARE_HEADER)

    main([str(fw_path)])
    stdout, _ = capsysbinary.readouterr()
    assert b"\xff\x00\xaa\x55FGVM64-7.04-FW-build2573-240201-" in stdout



from __future__ import annotations

import json
import re
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.helpers.utils import year_rollover_helper
from dissect.target.plugin import export
from dissect.target.plugins.apps.remoteaccess.remoteaccess import (
    GENERIC_FILE_TRANSFER_RECORD_FIELDS,
    GENERIC_LOG_RECORD_FIELDS,
    RemoteAccessPlugin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target


RE_TS = re.compile(r"(\w{3}\d{2}\s\d{2}:\d{2}:\d{2}\.\d{3})")


class SplashtopPlugin(RemoteAccessPlugin):
    """Splashtop plugin."""

    __namespace__ = "splashtop"

    RemoteAccessLogRecord = TargetRecordDescriptor(
        "remoteaccess/splashtop/log",
        GENERIC_LOG_RECORD_FIELDS,
    )

    RemoteAccessFileTransferRecord = TargetRecordDescriptor(
        "remoteaccess/splashtop/filetransfer",
        GENERIC_FILE_TRANSFER_RECORD_FIELDS,
    )

    LOG_PATHS = (
        # General agent log including connections and filetransfers
        "sysvol/Program Files (x86)/Splashtop/Splashtop Remote/Server/log/SPLog.txt",
        # File transfer log which is currently acquired but not parsed
        # All content is also included in the SPLog so no need to parse
        # acquired the file just in case something is missing in the SPLog
        # "sysvol/ProgramData/Splashtop/Temp/log/FTCLog.txt",
    )

    def __init__(self, target: Target):
        super().__init__(target)

        self.log_files: set[TargetPath] = set()

        for log_path in self.LOG_PATHS:
            if (log_file := self.target.fs.path(log_path)).exists():
                self.log_files.add(log_file)

    def check_compatible(self) -> None:
        if not self.log_files:
            raise UnsupportedPluginError("No Splashtop log files found on target")

    @export(record=RemoteAccessLogRecord)
    def logs(self) -> Iterator[RemoteAccessLogRecord]:
        """Parse Splashtop log files.

        Splashtop is a remote desktop application that can be used to get (persistent) access to a machine.
        It might be used in combination with Atera Management Agent.

        References:
            - https://www.synacktiv.com/en/publications/legitimate-rats-a-comprehensive-forensic-analysis-of-the-usual-suspects#atera-and-splashtop
        """
        target_tz = self.target.datetime.tzinfo

        for log_file in self.log_files:
            try:
                for ts, line in year_rollover_helper(log_file, RE_TS, "%b%d %H:%M:%S.%f", target_tz):
                    try:
                        # The line is of format "<#>%b%d %H:%M:%S.%f ..." check if the start matches an expected line
                        if (line[0], line[2]) != ("<", ">"):
                            self.target.log.error("LINE %s", line)
                            raise ValueError("Line does not match expected format")  # noqa: TRY301

                        # The prefix contains two spaces splitting off the timestamp, grab only the message part
                        message = line.split(" ", maxsplit=2)[-1]

                        yield self.RemoteAccessLogRecord(
                            ts=ts,
                            message=message,
                            source=log_file,
                            _target=self.target,
                        )
                    except ValueError as e:  # noqa: PERF203
                        self.target.log.warning("Could not parse log line in file %s: %r", log_file, line)
                        self.target.log.debug("", exc_info=e)
            except Exception as e:  # noqa: PERF203
                self.target.log.warning("Could not parse log file %s", log_file)
                self.target.log.debug("", exc_info=e)

    @export(record=RemoteAccessFileTransferRecord)
    def filetransfer(self) -> Iterator[RemoteAccessFileTransferRecord]:
        """Parse Splashtop filetransfers.

        Splashtop is a remote desktop application and can be used by adversaries to get (persistent) access to a
        machine. File transfers as logged in the generic logfile (``SPLog.txt``) show what files are downloaded
        to a system.
        """
        methods = ("OnUploadRequest", "OnUploadFileCPRequest", "OnDownloadRequest")
        for log_record in self.logs():
            try:
                # Example log entry:
                # SM_03280[FTCnnel] OnUploadFileCPRequest 1, 1 =>{"fileID":"353841253","fileName":"NOTE.txt","fileSize":"34","remotesessionFTC":1,"request":"uploadFile"}  # noqa: E501

                if any(method in log_record.message for method in methods):
                    json_data = json.loads(log_record.message.split("=>")[1])

                    yield self.RemoteAccessFileTransferRecord(
                        ts=log_record.ts,
                        message=log_record.message,
                        source=log_record.source,
                        filename=json_data["fileName"],
                        _target=self.target,
                    )
            except Exception as e:  # noqa: PERF203
                self.target.log.warning(
                    "Could not parse file transfer from message in file %s: %r", log_record.source, log_record.message
                )
                self.target.log.debug("", exc_info=e)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import ChildTargetRecord
from dissect.target.plugin import ChildTargetPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target import Target


class LimaChildTargetPlugin(ChildTargetPlugin):
    """Child target plugin that yields Lima VMs.

    Lima (Linux Machines) is a Linux VM or container runtime for macOS and Linux.

    References:
        - https://github.com/lima-vm/lima
    """

    __type__ = "lima"

    def __init__(self, target: Target):
        super().__init__(target)
        self.paths = []
        for user in self.target.user_details.all_with_home():
            # check .lima folder in ~/
            if (path := user.home_path.joinpath(".lima")).exists():
                self.paths.append(path)
            # check .lima folder in ~/.config/
            if (path := user.home_path.joinpath(".config", "lima")).exists():
                self.paths.append(path)

    def check_compatible(self) -> None:
        if not self.paths:
            raise UnsupportedPluginError("No Lima configurations found")

    def list_children(self) -> Iterator[ChildTargetRecord]:
        for path in self.paths:
            for instance in path.iterdir():
                if instance.name.startswith((".", "_")) or not instance.is_dir():
                    continue

                yield ChildTargetRecord(
                    type=self.__type__,
                    name=instance.name,
                    path=instance.joinpath("diffdisk"),
                    _target=self.target,
                )



from __future__ import annotations

import sys
from unittest.mock import MagicMock

import pytest

from dissect.target.exceptions import NotADirectoryError


def test_cb_filesystem_windows(monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        if "dissect.target.filesystems.cb" in sys.modules:
            m.delitem(sys.modules, "dissect.target.filesystems.cb")

        mock_cbc_sdk = MagicMock()
        m.setitem(sys.modules, "cbc_sdk", mock_cbc_sdk)
        m.setitem(sys.modules, "cbc_sdk.live_response_api", mock_cbc_sdk.live_response_api)

        from dissect.target.filesystems.cb import OS, CbFilesystem

        mock_session = MagicMock()
        mock_session.os_type = OS.WINDOWS
        mock_session.list_directory.return_value = [
            {
                "filename": "System32",
                "attributes": ["DIRECTORY"],
                "last_access_time": "2077-01-01T00:00:00Z",
                "last_write_time": "2022-10-04T00:00:00Z",
                "create_time": "2069-01-01T00:00:00Z",
                "size": 0,
            }
        ]

        fs = CbFilesystem(mock_session, "C:\\")

        root = fs.get("")
        mock_session.list_directory.assert_not_called()

        assert root.path == ""
        assert root.is_dir()
        assert not root.is_file()
        assert not root.is_symlink()

        entry = fs.get("windows/system32")
        mock_session.list_directory.assert_called_with("c:\\windows\\system32")

        assert entry.path == "windows/system32"
        assert entry.name == "system32"
        assert entry.is_dir()
        assert not entry.is_file()
        assert not entry.is_symlink()

        mock_session.list_directory.return_value = [
            {
                "filename": "myfile.txt",
                "attributes": ["ARCHIVE"],
                "last_access_time": "2077-01-01T00:00:00Z",
                "last_write_time": "2022-10-04T00:00:00Z",
                "create_time": "2069-01-01T00:00:00Z",
                "size": 0,
            }
        ]

        entries = list(entry.scandir())
        mock_session.list_directory.assert_called_with("c:\\windows\\system32\\")

        assert len(entries) == 1
        assert entries[0].path == "windows/system32/myfile.txt"
        assert entries[0].name == "myfile.txt"
        assert not entries[0].is_dir()
        assert entries[0].is_file()
        assert not entries[0].is_symlink()

        with pytest.raises(NotADirectoryError):
            entries[0].get().listdir()

        entries[0].get().open()
        mock_session.get_raw_file.assert_called_with("c:\\windows\\system32\\myfile.txt")

        stat_result = entries[0].stat()
        assert stat_result.st_mode == 0o100755
        assert stat_result.st_ino == 3013187826
        assert stat_result.st_atime == 3376684800.0
        assert stat_result.st_mtime == 1664841600.0
        assert stat_result.st_ctime == 3124224000.0



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

from dissect.cstruct import cstruct
from dissect.util import ts

from dissect.target.exceptions import FileNotFoundError, UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

LastLogRecord = TargetRecordDescriptor(
    "linux/log/lastlog",
    [
        ("datetime", "ts"),
        ("uint32", "uid"),
        ("string", "ut_user"),  # name
        ("string", "ut_host"),  # source
        ("string", "ut_tty"),  # port
    ],
)

lastlog_def = """
#define UT_NAMESIZE 32
#define UT_HOSTSIZE 256
#define size        292


struct {
    uint32 tv_sec;
} time_t;


struct entry {
    struct time_t ll_time;
    char    ut_user[UT_NAMESIZE];
    char    ut_host[UT_HOSTSIZE];
};
"""

c_lastlog = cstruct().load(lastlog_def)


class LastLogFile:
    def __init__(self, fh: BinaryIO):
        self.fh = fh

    def __iter__(self):
        while True:
            try:
                yield c_lastlog.entry(self.fh)
            except EOFError:  # noqa: PERF203
                break


class LastLogPlugin(Plugin):
    """Unix lastlog plugin."""

    def check_compatible(self) -> None:
        lastlog = self.target.fs.path("/var/log/lastlog")
        if not lastlog.exists():
            raise UnsupportedPluginError("No lastlog file found")

    @export(record=LastLogRecord)
    def lastlog(self) -> Iterator[LastLogRecord]:
        """Return last logins information from /var/log/lastlog.

        The lastlog file contains the most recent logins of all users on a Unix based operating system.

        References:
            - https://www.tutorialspoint.com/unix_commands/lastlog.htm
        """
        try:
            lastlog = self.target.fs.open("/var/log/lastlog")
        except FileNotFoundError:
            return

        users = {}
        for user in self.target.users():
            users[user.uid] = user.name

        log = LastLogFile(lastlog)

        for idx, entry in enumerate(log):
            # if ts=0 the uid has never logged in before
            if entry.ut_host.strip(b"\x00") == b"" or entry.ll_time.tv_sec == 0:
                continue

            yield LastLogRecord(
                ts=ts.from_unix(entry.ll_time.tv_sec),
                uid=idx,
                ut_user=users.get(idx),
                ut_tty=entry.ut_user.decode().strip("\x00"),
                ut_host=entry.ut_host.decode(errors="ignore").strip("\x00"),
                _target=self.target,
            )



from __future__ import annotations

import plistlib
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from dissect.target.helpers.record import IOSUserRecord
from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.os.unix.bsd.darwin._os import (
    DarwinPlugin,
    detect_macho_arch,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem, VirtualFilesystem
    from dissect.target.target import Target


class IOSPlugin(DarwinPlugin):
    """Apple iOS plugin.

    References:
        - https://developer.apple.com/library/archive/documentation/FileManagement/Conceptual/FileSystemProgrammingGuide/FileSystemOverview/FileSystemOverview.html
        - https://corp.digitalcorpora.org/corpora/mobile/iOS17/
    """

    SYSTEM = "/private/var/preferences/SystemConfiguration/preferences.plist"
    GLOBAL = "/private/var/mobile/Library/Preferences/.GlobalPreferences.plist"
    VERSION = "/System/Library/CoreServices/SystemVersion.plist"

    # /private/etc/master.passwd is a copy of /private/etc/passwd
    PASSWD_FILES = ("/private/etc/passwd",)

    def __init__(self, target: Target):
        super().__init__(target)

        self._config = Config.load(
            target.fs.path(self.SYSTEM),
            target.fs.path(self.GLOBAL),
            target.fs.path(self.VERSION),
        )

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        for fs in target.filesystems:
            if fs.exists("/private/var/preferences") and fs.exists("/private/var/mobile"):
                return fs
        return None

    @classmethod
    def create(cls, target: Target, sysvol: VirtualFilesystem) -> Self:
        target.fs.mount("/", sysvol)
        return cls(target)

    @export(property=True)
    def hostname(self) -> str | None:
        try:
            # ComputerName can contain invalid utf characters, so we use HostName instead.
            return self._config.SYSTEM["System"]["System"]["HostName"]
        except KeyError:
            pass

    @export(property=True)
    def ips(self) -> list:
        return []

    @export(property=True)
    def version(self) -> str:
        return f"{self._config.VERSION['ProductName']} {self._config.VERSION['ProductVersion']} ({self._config.VERSION['ProductBuildVersion']})"  # noqa: E501

    @export(record=IOSUserRecord)
    def users(self) -> Iterator[IOSUserRecord]:
        for user in super().users():
            yield IOSUserRecord(**user._asdict())

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.IOS.value

    @export(property=True)
    def architecture(self) -> str | None:
        if arch := detect_macho_arch(["/bin/df", "/bin/ps", "/sbin/fsck", "/sbin/mount"], fs=self.target.fs):
            return f"{arch}-ios"
        return None


@dataclass
class Config:
    SYSTEM: dict[str, Any]
    GLOBAL: dict[str, Any]
    VERSION: dict[str, Any]

    @classmethod
    def load(cls, *args: list[Path]) -> Self:
        plists = []
        for path in args:
            if path.is_file():
                plists.append(plistlib.load(path.open("rb")))
            else:
                plists.append({})
        return cls(*plists)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.child.parallels import ParallelsChildTargetPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_parallels_child_detection(target_macos_users: Target, fs_macos: VirtualFilesystem) -> None:
    """Test if we correctly find Parallels child VMs on MacOS targets."""

    config = absolute_path("_data/plugins/child/parallels/config.pvs")
    fs_macos.map_file("Users/dissect/Parallels/Windows 11.pvm/config.pvs", config)
    fs_macos.map_file("Users/dissect/Documents/Parallels/Windows 10.pvm/config.pvs", config)
    fs_macos.map_file("Users/Shared/Parallels/Windows 7.pvm/config.pvs", config)
    fs_macos.map_file(
        "Users/dissect/Library/Group Containers/svn.com.parallels.desktop.appstore/Shared/Parallels/My VM.pvm/config.pvs",  # noqa: E501
        config,
    )

    target_macos_users.add_plugin(ParallelsChildTargetPlugin)
    children = [child for _, child in target_macos_users.list_children()]

    assert len(children) == 4

    assert children[0].type == "parallels"
    assert children[0].name == "My VM"
    assert children[0].path == "/Users/Shared/Parallels/Windows 7.pvm"

    assert children[1].type == "parallels"
    assert children[1].name == "My VM"
    assert children[1].path == "/Users/dissect/Parallels/Windows 11.pvm"

    assert children[2].type == "parallels"
    assert children[2].name == "My VM"
    assert children[2].path == "/Users/dissect/Documents/Parallels/Windows 10.pvm"

    assert children[3].type == "parallels"
    assert children[3].name == "My VM"
    assert (
        children[3].path
        == "/Users/dissect/Library/Group Containers/svn.com.parallels.desktop.appstore/Shared/Parallels/My VM.pvm"
    )



from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.filesystem import VirtualFile, VirtualFilesystem
from dissect.target.plugins.filesystem.resolver import ResolverPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("test_target", "resolve_func"),
    [
        (
            "target_win",
            "resolve_windows",
        ),
        (
            "target_unix",
            "resolve_default",
        ),
    ],
)
def test_resolve(target_win: Target, target_unix: Target, test_target: str, resolve_func: str) -> None:
    targets = {
        "target_win": target_win,
        "target_unix": target_unix,
    }
    test_target = targets.get(test_target)
    resolver_plugin = ResolverPlugin(test_target)
    with patch.object(resolver_plugin, resolve_func, autospec=True):
        resolver_plugin.resolve("/some/path", "some_user")
        resolve_func = getattr(resolver_plugin, resolve_func)
        resolve_func.assert_called()


def test_resolve_no_path(target_win: Target) -> None:
    path = ""
    assert target_win.resolve(path) == target_win.fs.path(path)


def mock_expand_env(path: str, user_sid: str | None = None) -> str:
    return path


def mock_user_env(user: str | None) -> OrderedDict[str, str]:
    path_envs = {
        None: OrderedDict(
            (("%path%", "sysvol/windows/syswow64"),),
        ),
        "1337": OrderedDict(
            (("%path%", "sysvol\\some_dir\\;sysvol\\other_dir\\"),),
        ),
    }

    return path_envs.get(user, OrderedDict())


@pytest.mark.parametrize(
    ("path", "user", "resolved_path"),
    [
        (
            "",
            None,
            "",
        ),
        (
            "/systemroot/test",
            None,
            "/%systemroot%/test",
        ),
        (
            "\\systemroot\\test",
            None,
            "/%systemroot%/test",
        ),
        (
            "/??/C:/test",
            None,
            "C:/test",
        ),
        (
            "\\??\\C:\\test",
            None,
            "C:/test",
        ),
        (
            '"\\??\\C:\\windows\\system32\\calc.exe" -args',
            None,
            "C:/windows/system32/calc.exe",
        ),
        (
            "\\\\?\\C:\\Users\\Testpath\\Testfile",
            None,
            "C:/Users/Testpath/Testfile",
        ),
        (
            "C:/windows/system32/calc.exe -args",
            None,
            "C:/windows/system32/calc.exe",
        ),
        (
            "C:/some dir with spaces/calc.exe",
            None,
            "C:/some dir with spaces/calc.exe",
        ),
        (
            "C:/some dir with spaces/calc.exe -with -arguments",
            None,
            "C:/some dir with spaces/calc.exe",
        ),
        (
            "C:/windows/system32/calc -args",
            None,
            "C:/windows/system32/calc.exe",
        ),
        (
            "calc -args",
            None,
            "sysvol/windows/syswow64/calc.exe",
        ),
        (
            "calc",
            "no_env_use_fallback",
            "sysvol/windows/system32/calc.exe",
        ),
        (
            "calc",
            "1337",
            "sysvol/some_dir/calc.exe",
        ),
        (
            "foo",
            "1337",
            "sysvol/other_dir/foo.bat",
        ),
    ],
)
def test_resolve_windows(
    target_win: Target, path: str, fs_win: VirtualFilesystem, user: str | None, resolved_path: str
) -> None:
    fs_win.map_file_entry("windows/system32/calc.exe", VirtualFile(fs_win, "windows/system32/calc.exe", None))
    fs_win.map_file_entry("windows/syswow64/calc.exe", VirtualFile(fs_win, "windows/syswow64/calc.exe", None))
    fs_win.map_file_entry("some dir with spaces/calc.exe", VirtualFile(fs_win, "some dir with spaces/calc.exe", None))
    fs_win.map_file_entry("some_dir/calc.exe", VirtualFile(fs_win, "some_dir/calc.exe", None))
    fs_win.map_file_entry("other_dir/calc.exe", VirtualFile(fs_win, "other_dir/calc.exe", None))
    fs_win.map_file_entry("other_dir/foo.bat", VirtualFile(fs_win, "other_dir/foo.bat", None))

    resolver_plugin = ResolverPlugin(target_win)
    assert target_win.pathext is not None  # This will load the EnvironmentVariablePlugin
    env_plugin = next(plugin for plugin in target_win._plugins if type(plugin).__name__ == "EnvironmentVariablePlugin")
    with (
        patch.object(env_plugin, "_get_pathext", return_value={".exe", ".bat"}, autospec=True),
        patch.object(env_plugin, "expand_env", side_effect=mock_expand_env, autospec=True),
        patch.object(env_plugin, "user_env", side_effect=mock_user_env, autospec=True),
    ):
        assert resolver_plugin.resolve_windows(path, user_sid=user) == resolved_path


@pytest.mark.parametrize(
    ("path", "user", "resolved_path"),
    [
        (
            "",
            None,
            "",
        ),
        (
            "/foo/bar",
            None,
            "/foo/bar",
        ),
        (
            "/foo/",
            None,
            "/foo/",
        ),
        (
            "//foo//bar",
            1337,
            "/foo/bar",
        ),
    ],
)
def test_resolve_default(target_unix: Target, path: str, user: str, resolved_path: str) -> None:
    resolver_plugin = ResolverPlugin(target_unix)
    assert resolver_plugin.resolve_default(path, user_id=user) == resolved_path
